# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os, sys, time, base64
from deejayd.mediafilters import *
from deejayd.ui import log
from deejayd.database.querybuilders import *
from deejayd.database.dbobjects import SQLizer


############################################################################
class MediaFile(dict):

    def __init__(self, db, file_id):
        self.db = db
        self["media_id"] = file_id

    def set_info(self, key, value):
        self.set_infos({key: value})

    def set_infos(self, infos):
        self.db.set_media_infos(self["media_id"], infos)
        self.db.connection.commit()
        self.update(infos)

    def played(self):
        played = int(self["playcount"]) + 1
        timestamp = int(time.time())
        self.set_infos({"playcount":str(played), "lastplayed":str(timestamp)})

    def skip(self):
        skip = int(self["skipcount"]) + 1
        self.set_info("skipcount", str(skip))

    def get_cover(self):
        if self["type"] != "song":
            raise AttributeError
        try: (id, mime, cover) = self.db.get_file_cover(self["media_id"])
        except TypeError:
            raise AttributeError
        return {"cover": base64.b64decode(cover), "id":id, "mime": mime}

    def replay_gain(self):
        """Return the recommended Replay Gain scale factor."""
        try:
            db = float(self["replaygain_track_gain"].split()[0])
            peak = self["replaygain_track_peak"] and\
                     float(self["replaygain_track_peak"]) or 1.0
        except (KeyError, ValueError, IndexError):
            return 1.0
        else:
            scale = 10.**(db / 20)
            if scale * peak > 1:
                scale = 1.0 / peak # don't clip
            return min(15, scale)

############################################################################


class DatabaseQueries(object):
    structure_created = False

    def __init__(self, connection):
        self.connection = connection
        self.sqlizer = SQLizer()

    #
    # MediaDB requests
    #
    @query_decorator("lastid")
    def insert_file(self, cursor, dir, filename, lastmodified):
        query = "INSERT INTO library \
                (directory,name,lastmodified)VALUES(%s, %s, %s)"
        cursor.execute(query, (dir, filename, lastmodified))

    @query_decorator("none")
    def update_file(self, cursor, id, lastmodified):
        query = "UPDATE library SET lastmodified = %s WHERE id = %s"
        cursor.execute(query, (lastmodified, id))

    @query_decorator("rowcount")
    def set_media_infos(self, cursor, file_id, infos, allow_create = True):
        if allow_create:
            query = "REPLACE INTO media_info (id,ikey,value)VALUES(%s,%s,%s)"
            entries = [(file_id, k, v) for k, v in infos.items()]
        else:
            query = "UPDATE media_info SET value=%s WHERE id=%s and ikey=%s"
            entries = [(v, file_id, k) for k, v in infos.items()]
        cursor.executemany(query, entries)

    @query_decorator("none")
    def remove_file(self, cursor, id):
        queries = [
            "DELETE FROM library WHERE id = %s",
            "DELETE FROM media_info WHERE id = %s",
            "DELETE FROM medialist_libraryitem WHERE libraryitem_id = %s",
            ]
        for q in queries: cursor.execute(q, (id,))

    @query_decorator("fetchone")
    def is_file_exist(self, cursor, dirname, filename, type = "audio"):
        query = "SELECT d.id, l.id \
            FROM library l JOIN library_dir d ON d.id=l.directory\
            WHERE l.name = %s AND d.name = %s AND d.lib_type = %s"
        cursor.execute(query,(filename, dirname, type))

    @query_decorator("lastid")
    def insert_dir(self, cursor, new_dir, type="audio"):
        query = "INSERT INTO library_dir (name,type,lib_type)VALUES(%s,%s,%s)"
        cursor.execute(query, (new_dir, 'directory', type))

    @query_decorator("none")
    def remove_dir(self, cursor, id):
        cursor.execute("SELECT id FROM library WHERE directory = %s", (id,))
        for (id,) in cursor.fetchall():
            self.remove_file(id)
        cursor.execute("DELETE FROM library_dir WHERE id = %s", (id,))

    @query_decorator("none")
    def remove_recursive_dir(self, cursor, dir, type="audio"):
        files = self.get_all_files(dir, type)
        for file in files: self.remove_file(file[2])
        cursor.execute("DELETE FROM library_dir\
                      WHERE name LIKE %s AND lib_type = %s", (dir+u"%%", type))
        return [f[2] for f in files]

    @query_decorator("custom")
    def is_dir_exist(self, cursor, dirname, type):
        cursor.execute("SELECT id FROM library_dir\
            WHERE name=%s AND lib_type=%s", (dirname, type))
        rs = cursor.fetchone()
        return rs and rs[0]

    @query_decorator("none")
    def insert_dirlink(self, cursor, new_dirlink, type="audio"):
        query = "INSERT INTO library_dir (name,type,lib_type)VALUES(%s,%s,%s)"
        cursor.execute(query, (new_dirlink, "dirlink", type))

    @query_decorator("none")
    def remove_dirlink(self, cursor, dirlink, type="audio"):
        query = "DELETE FROM library_dir\
                 WHERE name = %s AND\
                       type = 'dirlink' AND\
                       lib_type = %s"
        cursor.execute(query, (dirlink, type))

    @query_decorator("fetchall")
    def get_dir_list(self, cursor, dir, t = "audio"):
        query = "SELECT DISTINCT id, name FROM library_dir\
                 WHERE name LIKE %s AND\
                       lib_type = %s AND\
                       type = 'directory'\
                 ORDER BY name"
        term = dir == unicode("") and u"%%" or dir+unicode("/%%")
        cursor.execute(query, (term, t))

    @query_decorator("fetchone")
    def get_file_info(self, cursor, file_id, info_type):
        query = "SELECT value FROM media_info WHERE id = %s AND ikey = %s"
        cursor.execute(query, (file_id, info_type))

    @query_decorator("fetchall")
    def get_all_files(self, cursor, dir, type = "audio"):
        query = "SELECT DISTINCT d.id, d.name, l.id, l.name, l.lastmodified\
            FROM library l JOIN library_dir d ON d.id=l.directory\
            WHERE d.name LIKE %s AND d.lib_type = %s ORDER BY d.name,l.name"
        cursor.execute(query,(dir+u"%%", type))

    @query_decorator("fetchall")
    def get_all_dirs(self, cursor, dir, type = "audio"):
        query = "SELECT DISTINCT id,name FROM library_dir\
            WHERE name LIKE %s AND type='directory' AND lib_type = %s\
            ORDER BY name"
        cursor.execute(query,(dir+u"%%", type))

    @query_decorator("fetchall")
    def get_all_dirlinks(self, cursor, dirlink, type = 'audio'):
        query = "SELECT DISTINCT id,name FROM library_dir\
            WHERE name LIKE %s AND type='dirlink' AND lib_type = %s\
            ORDER BY name"
        cursor.execute(query,(dirlink+u"%%", type))

    def _medialist_answer(self, answer, infos = []):
        files = []
        for m in answer:
            current = MediaFile(self, m[0])
            for index, attr in enumerate(infos):
                current[attr] = m[index+1]
            files.append(current)

        return files

    def _build_media_query(self, infos_list):
        selectquery = "i.id"
        joinquery = ""
        for index, key in enumerate(infos_list):
            selectquery += ",i%d.value" % index
            joinquery += " LEFT OUTER JOIN media_info i%d ON i%d.id=i.id AND\
                i%d.ikey='%s'" % (index, index, index, key)
        return selectquery, joinquery

    @query_decorator("medialist")
    def get_dir_content(self, cursor, dir, infos = [], type = "audio"):
        selectquery, joinquery = self._build_media_query(infos)
        query = "SELECT DISTINCT "+ selectquery +\
            " FROM library l JOIN library_dir d ON d.id=l.directory\
                           LEFT OUTER JOIN media_info i ON i.id=l.id"\
                           + joinquery+\
            " WHERE d.name = %s AND d.lib_type = %s ORDER BY d.name,l.name"
        cursor.execute(query,(dir, type))

    @query_decorator("medialist")
    def get_file(self, cursor, dir, file, infos = [], type = "audio"):
        selectquery, joinquery = self._build_media_query(infos)
        query = "SELECT DISTINCT "+ selectquery +\
            " FROM library l JOIN library_dir d ON d.id=l.directory\
                           JOIN media_info i ON i.id=l.id"\
                           + joinquery+\
            " WHERE d.name = %s AND l.name = %s AND d.lib_type = %s"
        cursor.execute(query, (dir, file, type))

    @query_decorator("medialist")
    def get_file_withids(self, cursor, file_ids, infos=[], type="audio"):
        selectquery, joinquery = self._build_media_query(infos)
        query = "SELECT DISTINCT "+ selectquery +\
            " FROM library l JOIN library_dir d ON d.id=l.directory\
                           JOIN media_info i ON i.id=l.id"\
                           + joinquery+\
            " WHERE l.id IN (%s) AND d.lib_type = '%s'" % \
                (",".join(map(str,file_ids)), type)
        cursor.execute(query)

    @query_decorator("medialist")
    def get_alldir_files(self, cursor, dir, infos = [], type = "audio"):
        selectquery, joinquery = self._build_media_query(infos)
        query = "SELECT DISTINCT "+ selectquery +\
            " FROM library l JOIN library_dir d ON d.id=l.directory\
                           JOIN media_info i ON i.id=l.id"\
                           + joinquery+\
            " WHERE d.name LIKE %s AND d.lib_type = %s ORDER BY d.name,l.name"
        cursor.execute(query,(dir+u"%%", type))

    @query_decorator("fetchall")
    def get_dircontent_id(self, cursor, dir, type):
        query = "SELECT l.id\
            FROM library l JOIN library_dir d ON l.directory = d.id\
            WHERE d.lib_type = %s AND d.name = %s"
        cursor.execute(query,(type, dir))

    @query_decorator("fetchall")
    def search_id(self, cursor, key, value):
        query = "SELECT DISTINCT id FROM media_info WHERE ikey=%s AND value=%s"
        cursor.execute(query,(key, value))

    @query_decorator("medialist")
    def search(self, cursor, filter, infos = [], orders = [], limit = None):
        filter = self.sqlizer.translate(filter)
        query = MediaSelectQuery()
        query.select_id()
        for tag in infos:
            query.select_tag(tag)
        filter.restrict(query)
        for (tag, direction) in orders:
            query.order_by_tag(tag, direction == "descending")
        query.set_limit(limit)
        cursor.execute(query.to_sql(), query.get_args())

    @query_decorator("fetchall")
    def list_tags(self, cursor, tag, filter):
        filter = self.sqlizer.translate(filter)
        query = MediaSelectQuery()
        query.select_tag(tag)
        filter.restrict(query)
        query.order_by_tag(tag)
        cursor.execute(query.to_sql(), query.get_args())

    @query_decorator("fetchall")
    def get_media_keys(self, cursor, type):
        query = "SELECT DISTINCT m.ikey FROM media_info m\
                JOIN media_info m2 ON m.id = m2.id\
                WHERE m2.ikey='type' AND m2.value=%s"
        cursor.execute(query, (type,))
#
# Post update action
#
    @query_decorator("none")
    def set_variousartist_tag(self, cursor, fid, file_info):
        query = "SELECT DISTINCT m.id,m.value,m3.value FROM media_info m\
                JOIN media_info m2 ON m.id = m2.id\
                JOIN media_info m3 ON m.id = m3.id\
                WHERE m.ikey='various_artist' AND m2.ikey='album'\
                AND m2.value=%s AND m3.ikey='artist'"
        cursor.execute(query, (file_info["album"],))
        try: (id, various_artist, artist) = cursor.fetchone()
        except TypeError: # first song of this album
            return
        else:
            need_update = False
            if various_artist == "__various__":
                need_update, ids = True, [(fid,)]
            elif artist != file_info["artist"]:
                need_update = True
                cursor.execute("SELECT id FROM media_info\
                        WHERE ikey='album' AND value=%s", (file_info["album"],))
                ids = cursor.fetchall()

        if need_update:
            cursor.executemany("UPDATE media_info SET value = '__various__'\
                    WHERE ikey='various_artist' AND id = %s", ids)

    @query_decorator("none")
    def erase_empty_dir(self, cursor, type = "audio"):
        cursor.execute("SELECT DISTINCT name FROM library_dir\
            WHERE lib_type=%s", (type,))
        for (dirname,) in cursor.fetchall():
            rs = self.get_all_files(dirname, type)
            if len(rs) == 0: # remove directory
                cursor.execute("DELETE FROM library_dir WHERE name = %s",\
                    (dirname,))

    @query_decorator("none")
    def update_stats(self, cursor, type = "audio"):
        # record mediadb stats
        query = "UPDATE stats SET value = \
          (SELECT COUNT(DISTINCT m.value) FROM media_info m JOIN media_info m2\
            ON m.id = m2.id WHERE m.ikey=%s AND m2.ikey='type' AND m2.value=%s)\
          WHERE name = %s"
        if type == "audio":
            values = [("uri", "song", "songs"),
                    ("artist", "song", "artists"),
                    ("genre", "song", "genres"),
                    ("album", "song", "albums")]
        elif type == "video":
            values = [("uri", "video", "videos")]
        cursor.executemany(query, values)

        # update last updated stats
        cursor.execute("UPDATE stats SET value = %s WHERE name = %s",\
            (time.time(),type+"_library_update"))

    #
    # cover requests
    #
    @query_decorator("fetchone")
    def get_file_cover(self, cursor, file_id, source = False):
        var = source and "source" or "image"
        query = "SELECT c.id, c.mime_type, c." + var +\
            " FROM media_info m JOIN cover c\
                              ON m.ikey = 'cover' AND m.value = c.id\
            WHERE m.id = %s"
        cursor.execute(query, (file_id,))

    @query_decorator("fetchone")
    def is_cover_exist(self, cursor, source):
        query = "SELECT id,lmod FROM cover WHERE source=%s"
        cursor.execute(query, (source,))

    @query_decorator("lastid")
    def add_cover(self, cursor, source, mime, image):
        query = "INSERT INTO cover (source,mime_type,lmod,image)\
                VALUES(%s,%s,%s,%s)"
        cursor.execute(query, (source, mime, time.time(), image))

    @query_decorator("none")
    def update_cover(self, cursor, id, mime, new_image):
        query = "UPDATE cover SET mime_type = %s, lmod = %s, image = %s\
                WHERE id=%s"
        cursor.execute(query, (mime, time.time(), new_image, id))

    @query_decorator("none")
    def remove_cover(self, cursor, id):
        query = "DELETE FROM cover WHERE id=%s"
        cursor.execute(query, (id,))

    @query_decorator("none")
    def remove_unused_cover(self, cursor):
        query = "DELETE FROM cover WHERE id NOT IN \
                  (SELECT DISTINCT value FROM media_info WHERE ikey = 'cover')"
        cursor.execute(query)

    #
    # common medialist requests
    #
    @query_decorator("fetchall")
    def get_medialist_list(self, cursor):
        query = SimpleSelect('medialist')
        query.select_column('id', 'name', 'type')
        query.order_by('name')
        cursor.execute(query.to_sql(), query.get_args())

    @query_decorator("custom")
    def get_medialist_id(self, cursor, pl_name, pl_type = 'static'):
        query = SimpleSelect('medialist')
        query.select_column('id')
        query.append_where("name = %s", (pl_name, ))
        query.append_where("type = %s", (pl_type, ))
        cursor.execute(query.to_sql(), query.get_args())

        ans = cursor.fetchone()
        if ans is None: raise ValueError
        return ans[0]

    @query_decorator("fetchone")
    def is_medialist_exists(self, cursor, pl_id):
        query = SimpleSelect('medialist')
        query.select_column('id', 'name', 'type')
        query.append_where("id = %s", (pl_id, ))
        cursor.execute(query.to_sql(), query.get_args())

    @query_decorator("none")
    def delete_medialist(self, cursor, ml_id):
        try: ml_id, name, type = self.is_medialist_exists(ml_id)
        except TypeError:
            return
        if type == "static":
            query = "DELETE FROM medialist_libraryitem WHERE medialist_id = %s"
            cursor.execute(query, (ml_id,))
        elif type == "magic":
            for (filter_id,) in self.__get_medialist_filterids(cursor, ml_id):
                self.delete_filter(cursor, filter_id)
            cursor.execute(\
              "DELETE FROM medialist_filters WHERE medialist_id=%s", (ml_id,))
            # delete medialist properties
            cursor.execute(\
              "DELETE FROM medialist_property WHERE medialist_id=%s", (ml_id,))
            # delete medialist sort
            cursor.execute(\
              "DELETE FROM medialist_sorts WHERE medialist_id=%s", (ml_id,))
        cursor.execute("DELETE FROM medialist WHERE id = %s", (ml_id,))
        self.connection.commit()

    def get_filter(self, cursor, id):
        try: filter_type = self.__get_filter_type(cursor, id)
        except ValueError:
            return None
        if filter_type == 'basic':
            return self.__get_basic_filter(cursor, id)
        elif filter_type == 'complex':
            return self.__get_complex_filter(cursor, id)

    def delete_filter(self, cursor, filter_id):
        try: filter_type = self.__get_filter_type(cursor, filter_id)
        except ValueError:
            return None

        if filter_type == 'basic':
            cursor.execute("DELETE FROM filters_basicfilters\
                WHERE filter_id = %s", (filter_id,))
        elif filter_type == 'complex':
            # get filters id associated with this filter
            query = SimpleSelect('filters_complexfilters_subfilters')
            query.select_column('filter_id')
            query.append_where("complexfilter_id = %s", (filter_id, ))
            cursor.execute(query.to_sql(), query.get_args())

            for (id,) in cursor.fetchall():
                self.delete_filter(cursor, id)
                cursor.execute("DELETE FROM filters_complexfilters_subfilters \
                    WHERE complexfilter_id = %s AND filter_id = %s",\
                    (filter_id, id))
            cursor.execute("DELETE FROM filters_complexfilters \
                WHERE filter_id = %s",(filter_id,))

        cursor.execute("DELETE FROM filters WHERE filter_id = %s",(filter_id,))

    def __get_filter_type(self, cursor, filter_id):
        query = SimpleSelect('filters')
        query.select_column('type')
        query.append_where("filter_id = %s", (filter_id, ))
        cursor.execute(query.to_sql(), query.get_args())
        record = cursor.fetchone()

        if not record: raise ValueError
        return record[0]

    def __get_basic_filter(self, cursor, id):
        query = SimpleSelect('filters_basicfilters')
        query.select_column('tag', 'operator', 'pattern')
        query.append_where("filter_id = %s", (id, ))
        cursor.execute(query.to_sql(), query.get_args())
        record = cursor.fetchone()

        if record:
            bfilter_class = NAME2BASIC[record[1]]
            f = bfilter_class(record[0], record[2])
            return f

    def __get_complex_filter(self, cursor, id):
        query = SimpleSelect('filters_complexfilters')
        query.select_column('combinator')
        query.append_where("filter_id = %s", (id, ))
        cursor.execute(query.to_sql(), query.get_args())
        record = cursor.fetchone()

        if record:
            cfilter_class = NAME2COMPLEX[record[0]]
            query = SimpleSelect('filters_complexfilters_subfilters')
            query.select_column('filter_id')
            query.append_where("complexfilter_id = %s", (id, ))
            cursor.execute(query.to_sql(), query.get_args())
            sf_records = cursor.fetchall()
            filterlist = []
            for sf_record in sf_records:
                sf_id = sf_record[0]
                filterlist.append(self.get_filter(cursor, sf_id))
            cfilter = cfilter_class(*filterlist)
            return cfilter

    def __get_medialist_filterids(self, cursor, ml_id):
        query = SimpleSelect('medialist_filters')
        query.select_column('filter_id')
        query.append_where("medialist_id = %s", (ml_id, ))
        cursor.execute(query.to_sql(), query.get_args())

        return cursor.fetchall()

    def __add_medialist_filters(self, cursor, pl_id, filters):
        filter_ids = []
        for filter in filters:
            filter_id = self.sqlizer.translate(filter).save(self)
            if filter_id: filter_ids.append((pl_id, filter_id))
        cursor.executemany("INSERT INTO medialist_filters\
            (medialist_id,filter_id)VALUES(%s,%s)", filter_ids)

    @query_decorator("custom")
    def get_magic_medialist_filters(self, cursor, ml_id):
        rs = self.__get_medialist_filterids(cursor, ml_id)
        if not rs: return []
        filters = []
        for (filter_id,) in rs:
            filter = self.get_filter(cursor, filter_id)
            if filter: filters.append(filter)
        return filters

    @query_decorator("custom")
    def set_magic_medialist_filters(self, cursor, pl_name, filters):
        slt_query = "SELECT id FROM medialist WHERE name=%s AND type = 'magic'"
        cursor.execute(slt_query, (pl_name,))
        rs = cursor.fetchone()
        if not rs:
            query = "INSERT INTO medialist (name,type)VALUES(%s,'magic')"
            cursor.execute(query, (pl_name,))
            id = self.connection.get_last_insert_id(cursor)
        else: (id,) = rs

        for (filter_id,) in self.__get_medialist_filterids(cursor, id):
            self.delete_filter(cursor, filter_id)
        cursor.execute(\
          "DELETE FROM medialist_filters WHERE medialist_id=%s", (id,))

        self.__add_medialist_filters(cursor, id, filters)
        self.connection.commit()
        return id

    @query_decorator("none")
    def add_magic_medialist_filters(self, cursor, pl_id, filters):
        self.__add_medialist_filters(cursor, pl_id, filters)
        self.connection.commit()

    @query_decorator("fetchall")
    def get_magic_medialist_sorts(self, cursor, ml_id):
        query = "SELECT tag,direction FROM medialist_sorts\
                WHERE medialist_id = %s"
        cursor.execute(query, (ml_id,))

    @query_decorator("none")
    def set_magic_medialist_sorts(self, cursor, ml_id, sorts):
        # first, delete all previous sort for this medialist
        cursor.execute("DELETE FROM medialist_sorts WHERE medialist_id=%s",\
                (ml_id,))
        cursor.executemany("INSERT INTO medialist_sorts\
            (medialist_id,tag,direction)VALUES(%s,%s,%s)",\
            [ (ml_id, tag, direction) for (tag, direction) in sorts])
        self.connection.commit()

    @query_decorator("fetchall")
    def get_magic_medialist_properties(self, cursor, ml_id):
        query = "SELECT ikey,value FROM medialist_property\
                WHERE medialist_id = %s"
        cursor.execute(query, (ml_id,))

    @query_decorator("none")
    def set_magic_medialist_property(self, cursor, ml_id, key, value):
        cursor.execute("REPLACE INTO medialist_property\
            (medialist_id,ikey,value)VALUES(%s,%s,%s)", (ml_id, key, value))
        self.connection.commit()

######################################
###### Static medialist queries ######
######################################
    @query_decorator("none")
    def add_to_static_medialist(self, cursor, ml_id, media_ids):
        query = "INSERT INTO medialist_libraryitem\
            (medialist_id, libraryitem_id) VALUES(%s,%s)"
        cursor.executemany(query, [(ml_id, mid) for mid in media_ids])

    @query_decorator("medialist")
    def get_static_medialist(self, cursor, ml_id, infos = []):
        selectquery, joinquery = self._build_media_query(infos)
        query = "SELECT DISTINCT "+ selectquery + ", mi.position " +\
            " FROM medialist m JOIN medialist_libraryitem mi\
                                    ON m.id = mi.medialist_id\
                           JOIN media_info i ON i.id=mi.libraryitem_id"\
                           + joinquery+\
            " WHERE m.id = %s AND m.type = 'static' ORDER BY mi.position"
        cursor.execute(query,(ml_id,))

    @query_decorator("custom")
    def set_static_medialist(self, cursor, name, content):
        slt_query = "SELECT id FROM medialist WHERE name=%s AND type = 'static'"
        cursor.execute(slt_query, (name,))
        rs = cursor.fetchone()
        if not rs:
            query = "INSERT INTO medialist (name,type)VALUES(%s,'static')"
            cursor.execute(query, (name,))
            id = self.connection.get_last_insert_id(cursor)
        else: (id,) = rs

        cursor.execute(\
            "DELETE FROM medialist_libraryitem WHERE medialist_id = %s",(id,))
        values = [(id, s["media_id"]) for s in content]
        query = "INSERT INTO medialist_libraryitem(medialist_id,libraryitem_id)\
            VALUES(%s,%s)"
        cursor.executemany(query,values)
        self.connection.commit()
        # return id of the playlist
        return id

    #
    # Webradio requests
    #
    @query_decorator("fetchall")
    def get_webradios(self, cursor):
        cursor.execute("SELECT DISTINCT w.id, w.name, e.url\
            FROM webradio w INNER JOIN webradio_entries e \
                            ON w.id = e.webradio_id\
            ORDER BY w.id, e.id")

    @query_decorator("none")
    def add_webradio(self, cursor, name, urls):
        query = "INSERT INTO webradio(name)VALUES(%s)"
        cursor.execute(query, (name,))
        wid = self.connection.get_last_insert_id(cursor)

        query = "INSERT INTO webradio_entries(url, webradio_id)VALUES(%s,%s)"
        cursor.executemany(query, [(url,wid) for url in urls])
        self.connection.commit()

    @query_decorator("none")
    def remove_webradios(self, cursor, wids):
        wids = [(wid,) for wid in wids]
        cursor.executemany("DELETE FROM webradio WHERE id = %s" , wids)
        cursor.executemany("DELETE FROM webradio_entries\
            WHERE webradio_id = %s", wids)
        self.connection.commit()

    @query_decorator("none")
    def remove_url_from_webradio(self, cursor, wid, url_id):
        cursor.execute("DELETE FROM webradio_entries\
            WHERE webradio_id = %s AND id = %s", (wid, url_id))
        self.connection.commit()

    @query_decorator("none")
    def add_url_for_webradio(self, cursor, wid, url):
        cursor.execute("INSERT INTO webradio_entries (url, webradio_id)\
            VALUES(%s,%s)", (url, wid))
        self.connection.commit()

    @query_decorator("none")
    def clear_webradios(self, cursor):
        cursor.execute("DELETE FROM webradio")
        cursor.execute("DELETE FROM webradio_entries")
        self.connection.commit()

    #
    # Stat requests
    #
    @query_decorator("fetchall")
    def get_stats(self, cursor):
        cursor.execute("SELECT * FROM stats")

    #
    # State requests
    #
    @query_decorator("none")
    def set_state(self, cursor, values):
        cursor.executemany("REPLACE INTO variables (value,name)VALUES(%s,%s)",\
                           values)
        self.connection.commit()

    @query_decorator("custom")
    def get_state(self, cursor, type):
        cursor.execute("SELECT value FROM variables WHERE name = %s",(type,))
        try:
            (rs,) =  cursor.fetchone()
            return rs
        except (ValueError, TypeError):
            return None

    def close(self):
        self.connection.close()

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.mediafilters import *
from deejayd.sources._base import _BaseSortedLibSource, SourceError
from deejayd.ui import log

class PanelSource(_BaseSortedLibSource):
    SUBSCRIPTIONS = {
            "playlist.update": "cb_playlist_update",
            "playlist.listupdate": "cb_playlist_listupdate",
            "mediadb.mupdate": "cb_library_changes",
            }
    base_medialist = "__panelcurrent__"
    name = "panel"
    source_signal = 'panel.update'
    supported_panel_tags = [\
            ['genre','artist','album'],\
            ['genre','various_artist','album'],\
            ['artist','album'],\
            ['various_artist','album'],\
            ]
    contains_tags = ('genre','artist','album','title','all')
    sort_tags = ('genre','artist','album','title','rating','tracknumber')
    default_sorts = DEFAULT_AUDIO_SORT
    _default_state = {"id": 1, "playorder": "inorder", "repeat": False,\
                      "panel-type": "panel", "panel-value": "0"}

    def __init__(self, db, library, config):
        super(PanelSource, self).__init__(db, library)

        # get panel tags
        self.__panel_tags = config.getlist("panel", "panel_tags")
        if self.__panel_tags not in self.supported_panel_tags:
            log.err(_("You choose wrong panel tags, fallback to default"))
            self.__panel_tags = ['genre','artist','album']

        # get recorded panel medialist
        filter = And()
        try: ml_id = self.db.get_medialist_id(self.base_medialist, 'magic')
        except ValueError: # medialist does not exist
            self._sorts = []
        else:
            # get filters
            filter.filterlist = self.db.get_magic_medialist_filters(ml_id)
            # get recorded sorts
            self._sorts = list(self.db.get_magic_medialist_sorts(ml_id)) or []
        self.__filters_to_parms(filter)

        # custom attributes
        self.__update_active_list(self._state["panel-type"],\
                                  self._state["panel-value"])

    def __filters_to_parms(self, filter = None):
        # filter as AND(Search, Panel) with
        #  * Panel : And(OR(genre=value1, genre=value2), OR(artist=value1)...)
        #  * Search : OR(tag1 CONTAINS value, )
        self.__search, self.__panel, self.__filter = None, {}, filter
        if filter != None:
            for ft in filter.filterlist:
                if ft.get_name() == "and": # panel
                    for panel_ft in ft.filterlist:
                        tag = panel_ft.filterlist[0].tag
                        if tag in self.__panel_tags:
                            self.__panel[tag] = panel_ft
                elif ft.get_name() == "or" or ft.type == "basic": # search
                    self.__search = ft

    def __update_panel_filters(self):
        # rebuild filter
        self.__filter = And()
        if self.__search:
            self.__filter.filterlist.append(self.__search)
        panel_filter = And()
        panel_filter.filterlist = self.__panel.values()
        self.__filter.filterlist.append(panel_filter)

        if self._state["panel-type"] == "panel":
            sorts = self._sorts + self.__class__.default_sorts
            medias = self.library.search(self.__filter, sorts)
            self._media_list.set(medias)
            self.__update_current()
            self.dispatch_signame(self.__class__.source_signal)

    def __update_active_list(self, type, pl_id, raise_ex = False):
        need_sort, sorts = False, self._sorts + self.__class__.default_sorts
        if type == "playlist":
            try: medias = self._get_playlist_content(pl_id)
            except SourceError: # playlist does not exist, set to panel
                if raise_ex:
                    raise SourceError(_("Playlist with id %s not found")\
                            % str(pl_id))
                self._state["panel-type"] = "panel";
                medias = self.library.search(self.__filter, sorts)
            else:
                need_sort = True
        elif type == "panel":
            medias = self.library.search(self.__filter, sorts)
        else:
            raise TypeError
        self._media_list.set(medias)
        if need_sort: self._media_list.sort(self._sorts)

    def __update_current(self):
        if self._current and self._current["id"] != -1: # update current id
            media_id = self._current["media_id"]
            try:
                self._current["id"] = self._media_list.find_id(media_id)
            except ValueError:
                self._current["id"] = -1

    def get_panel_tags(self):
        return self.__panel_tags

    def set_panel_filters(self, tag, values):
        if tag not in self.__panel_tags:
            raise SourceError(_("Tag '%s' not supported") % tag)
        if not values:
            self.remove_panel_filters(tag)
            return
        filter = Or()
        for value in values:
            filter.combine(Equals(tag, value))
        if tag in self.__panel and self.__panel[tag].equals(filter):
            return # do not need update
        self.__panel[tag] = filter

        # remove filter for panels at the right of this tag
        for tg in reversed(self.get_panel_tags()):
            if tg == tag: break
            try: del self.__panel[tg]
            except KeyError:
                pass

        self.__update_panel_filters()

    def remove_panel_filters(self, tag):
        try: del self.__panel[tag]
        except KeyError:
            pass
        self.__update_panel_filters()

    def clear_panel_filters(self):
        self.__panel = {}
        self.__update_panel_filters()

    def set_search_filter(self, tag, value):
        if tag not in self.__class__.contains_tags:
            raise SourceError(_("Tag '%s' not supported") % tag)
        if tag == "all":
            new_filter = Or()
            for tg in ('title','genre','artist','album'):
                new_filter.combine(Contains(tg, value))
        else:
            new_filter = Contains(tag, value)
        self.__search = new_filter
        self.__panel = {} # remove old panel filter
        self.__update_panel_filters()

    def clear_search_filter(self):
        if self.__search:
            self.__search = None
            self.__update_panel_filters()

    def get_active_list(self):
        return {"type": self._state["panel-type"],\
                 "value": self._state["panel-value"]}

    def set_active_list(self, type, pl_id):
        if type == self._state["panel-type"]\
                and pl_id == self._state["panel-value"]:
            return # we do not need to update panel
        self.__update_active_list(type, pl_id, raise_ex = True)
        self._state["panel-type"] = type
        self._state["panel-value"] = pl_id
        self.__update_current()
        self.dispatch_signame(self.__class__.source_signal)

    def get_content(self, start = 0, stop = None):
        if self._state["panel-type"] == "panel":
            return self._media_list.get(start, stop),self.__filter,self._sorts
        elif self._state["panel-type"] == "playlist":
            return self._media_list.get(start, stop), None, self._sorts

    def close(self):
        super(PanelSource, self).close()
        # save panel filters
        filter_list = self.__filter.filterlist
        ml_id = self.db.set_magic_medialist_filters(self.base_medialist,\
                filter_list)
        # save panel sorts
        self.db.set_magic_medialist_sorts(ml_id, self._sorts)

    #
    # callback for deejayd signal
    #
    def cb_playlist_update(self, signal):
        pl_id = int(signal.get_attr('pl_id'))
        if self._state["panel-type"] == "playlist"\
                and pl_id == int(self._state["panel-value"]):
            self.__update_active_list("playlist", pl_id, raise_ex = True)
            self.dispatch_signame(self.__class__.source_signal)

    def cb_playlist_listupdate(self, signal):
        if self._state["panel-type"] == "playlist":
            pl_id = int(self._state["panel-value"])
            list = [int(id) \
                    for (id, pl, type) in self.db.get_medialist_list() if not \
                    pl.startswith("__") or not pl.endswith("__")]
            if pl_id not in list: # fall back to panel
                self.__update_active_list("panel", "", raise_ex = True)
                self._state["panel-type"] = "panel"
                self._state["panel-value"] = ""
                self.dispatch_signame(self.__class__.source_signal)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from deejayd.mediadb.library import NotFoundException
from deejayd.sources._base import _BaseSortedLibSource, SourceError
from deejayd import mediafilters

class VideoSource(_BaseSortedLibSource):
    SUBSCRIPTIONS = {
            "mediadb.mupdate": "cb_library_changes",
            }
    name = "video"
    base_medialist = "__videocurrent__"
    source_signal = 'video.update'
    sort_tags = ('title','rating','length')
    default_sorts = mediafilters.DEFAULT_VIDEO_SORT

    def __init__(self, db, library):
        super(VideoSource, self).__init__(db, library)
        # load saved
        try: ml_id = self.db.get_medialist_id(self.base_medialist, 'static')
        except ValueError: # medialist does not exist
            self._sorts = []
        else:
            self._sorts = list(self.db.get_magic_medialist_sorts(ml_id)) or []
            self._media_list.set(self._get_playlist_content(ml_id))
            self.set_sorts(self._sorts)

    def set(self, type, value):
        need_sort = False
        if type == "directory":
            try: video_list = self.library.get_all_files(value)
            except NotFoundException:
                raise SourceError(_("Directory %s not found") % value)
            need_sort = True
        elif type == "search":
            sorts = self._sorts + self.__class__.default_sorts
            video_list = self.library.search(\
                mediafilters.Contains("title",value), sorts)
        else:
            raise SourceError(_("type %s not supported") % type)

        self._media_list.set(video_list)
        if need_sort:
            self._media_list.sort(self._sorts + self.default_sorts)
        self.dispatch_signame(self.source_signal)

    def get_content(self, start = 0, stop = None):
        return self._media_list.get(start, stop), None, self._sorts

    def close(self):
        super(VideoSource, self).close()
        # save panel sorts
        try: ml_id = self.db.get_medialist_id(self.base_medialist, 'static')
        except ValueError: # medialist does not exist
            pass
        else:
            self.db.set_magic_medialist_sorts(ml_id, self._sorts)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import sys
import ctypes


try:
    _libX11 = ctypes.cdll.LoadLibrary('libX11.so.6')
except (ImportError, OSError), e:
    raise ImportError, e


# int XInitThreads()
_libX11.XInitThreads.restype = ctypes.c_int

# Display *XOpenDisplay(char *display_name)
_libX11.XOpenDisplay.restype = ctypes.c_void_p
_libX11.XOpenDisplay.argtypes = (ctypes.c_char_p, )

# int XDefaultScreen(Display *display)
_libX11.XDefaultScreen.restype = ctypes.c_int
_libX11.XDefaultScreen.argtypes = (ctypes.c_void_p, )

# int XDisplayWidth(Display *display, int screen_number);
_libX11.XDisplayWidth.restype = ctypes.c_int
_libX11.XDisplayWidth.argtypes = (ctypes.c_void_p, ctypes.c_int)

# int XDisplayHeight(Display *display, int screen_number);
_libX11.XDisplayHeight.restype = ctypes.c_int
_libX11.XDisplayHeight.argtypes = (ctypes.c_void_p, ctypes.c_int)

# int XDisplayWidthMM(Display *display, int screen_number);
_libX11.XDisplayWidthMM.restype = ctypes.c_int
_libX11.XDisplayWidthMM.argtypes = (ctypes.c_void_p, ctypes.c_int)

# int XDisplayHeightMM(Display *display, int screen_number);
_libX11.XDisplayHeightMM.restype = ctypes.c_int
_libX11.XDisplayHeightMM.argtypes = (ctypes.c_void_p, ctypes.c_int)

# void XCloseDisplay(Display *display)
_libX11.XCloseDisplay.argtypes = (ctypes.c_void_p, )

# void XLockDisplay(Display *display)
_libX11.XLockDisplay.argtypes = (ctypes.c_void_p, )

# void XUnlockDisplay(Display *display)
_libX11.XUnlockDisplay.argtypes = (ctypes.c_void_p, )

# Window XDefaultRootWindow(Display *display)
_libX11.XDefaultRootWindow.restype = ctypes.c_ulong
_libX11.XDefaultRootWindow.argtypes = (ctypes.c_void_p, )

# Window XCreateSimpleWindow(Display *display, Window parent, int x, int y,
#                            uint width, uint height, uint border_width,
#                            ulong border, ulong background)
_libX11.XCreateSimpleWindow.restype = ctypes.c_ulong
_libX11.XCreateSimpleWindow.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
                                        ctypes.c_int, ctypes.c_int,
                                        ctypes.c_uint, ctypes.c_uint,
                                        ctypes.c_uint, ctypes.c_ulong,
                                        ctypes.c_ulong)

# void XMapRaised(Display *display, Window w)
_libX11.XMapRaised.argtypes = (ctypes.c_void_p, ctypes.c_ulong)

# void XUnmapWindow(Display *display, Window w)
_libX11.XUnmapWindow.argtypes = (ctypes.c_void_p, ctypes.c_ulong)

# void XSelectInput(Display *display, Window w, long event_mask)
_libX11.XSelectInput.argtypes = (ctypes.c_void_p, ctypes.c_ulong, ctypes.c_long)

# int XGetGeometry(Display *display, Drawable d, Window *root_return,
#                  int *x_return, int *y_return, uint *width_return,
#                  uint *height_return, uint *border_width_return,
#                  uint *depth_return)
_libX11.XGetGeometry.restype = ctypes.c_int
_libX11.XGetGeometry.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
    ctypes.POINTER(ctypes.c_ulong), ctypes.POINTER(ctypes.c_int),
    ctypes.POINTER(ctypes.c_int), ctypes.POINTER(ctypes.c_uint),
    ctypes.POINTER(ctypes.c_uint), ctypes.POINTER(ctypes.c_uint),
    ctypes.POINTER(ctypes.c_uint))

# void XSync(Display *display, Bool discard)
_libX11.XSync.argtypes = (ctypes.c_void_p, ctypes.c_int)

# void XDestroyWindow(Display *display, Window w)
_libX11.XDestroyWindow.argtypes = (ctypes.c_void_p, ctypes.c_ulong)

class XColor(ctypes.Structure):
    _fields_ = (
                 ('pixel', ctypes.c_ulong),
                 ('red', ctypes.c_ushort),
                 ('green', ctypes.c_ushort),
                 ('blue', ctypes.c_ushort),
                 ('flags', ctypes.c_char),
                 ('pad', ctypes.c_char),
               )

class MWMHints(ctypes.Structure):
    MWM_HINTS_DECORATIONS   = (1L << 1)
    PROP_MWM_HINTS_ELEMENTS = 5

    _fields_ = (
                 ('flags', ctypes.c_uint),
                 ('functions', ctypes.c_uint),
                 ('decorations', ctypes.c_uint),
                 ('input_mode', ctypes.c_int),
                 ('status', ctypes.c_uint),
               )

# Atom XInternAtom(Display *display,char *atom_name, Bool only_if_exists)
_libX11.XInternAtom.restype = ctypes.c_ulong
_libX11.XInternAtom.argtypes = (ctypes.c_void_p, ctypes.c_char_p, ctypes.c_int)

# XChangeProperty(Display *display, Window w, Atom property, Atom type,
#                 int format, int mode, uchar *data, int nelements)
_libX11.XChangeProperty.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
                                    ctypes.c_ulong, ctypes.c_ulong,
           ctypes.c_int, ctypes.c_int, ctypes.c_char_p, ctypes.c_int)

class PropertyModes:
    PropModeReplace = 0
    PropModePrepend = 1
    PropModeAppend = 2

class XSetWindowAttributes(ctypes.Structure):
    _fields_ = (
                 ('background_pixmap', ctypes.c_ulong),
                 ('background_pixel', ctypes.c_ulong),
                 ('border_pixmap', ctypes.c_ulong),
                 ('border_pixel', ctypes.c_ulong),
                 ('bit_gravity', ctypes.c_int),
                 ('win_gravity', ctypes.c_int),
                 ('backing_store', ctypes.c_int),
                 ('backing_planes', ctypes.c_ulong),
                 ('backing_pixel', ctypes.c_ulong),
                 ('save_under', ctypes.c_int),
                 ('event_mask', ctypes.c_long),
                 ('do_not_propagate_mask', ctypes.c_long),
                 ('override_redirect', ctypes.c_int),
                 ('colormap', ctypes.c_ulong),
                 ('cursor', ctypes.c_ulong),
               )

# XChangeWindowAttributes(Display *display, Window w, ulong valuemask,
#                         XSetWindowAttributes *attributes)
_libX11.XChangeWindowAttributes.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
                                    ctypes.c_ulong,
                                    ctypes.POINTER(XSetWindowAttributes))

CWOverrideRedirect = (1L<<9)

# Colormap XDefaultColormap(Display *display, int screen)
_libX11.XDefaultColormap.restype = ctypes.c_ulong
_libX11.XDefaultColormap.argtypes = (ctypes.c_void_p, ctypes.c_int)

# int XAllocNamedColor(Display *display, Colormap cmap, char *color_name,
#                      XColor *screen_def_return, XColor *exact_def_return)
_libX11.XAllocNamedColor.restype = ctypes.c_int
_libX11.XAllocNamedColor.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
                                     ctypes.c_char_p, ctypes.POINTER(XColor),
                                     ctypes.POINTER(XColor))

# XDefineCursor(Display *display, Window w, Cursor cursor)
_libX11.XDefineCursor.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
                                  ctypes.c_ulong)

# XUndefineCursor(Display *display, Window w)
_libX11.XUndefineCursor.argtypes = (ctypes.c_void_p, ctypes.c_ulong)

# Cursor XCreatePixmapCursor(Display *display, Pixmap source, Pixmap mask,
#                            XColor *foreground_color,
#                            XColor *background_color, uint x, uint y)
_libX11.XCreatePixmapCursor.restype = ctypes.c_ulong
_libX11.XCreatePixmapCursor.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
    ctypes.c_ulong, ctypes.POINTER(XColor),
    ctypes.POINTER(XColor), ctypes.c_uint, ctypes.c_uint)

# XFreeCursor(Display *display, Cursor cursor)
_libX11.XFreeCursor.argtypes = (ctypes.c_void_p, ctypes.c_ulong)

# Colormap XDefaultColormap(Display *display, int screen)
_libX11.XDefaultColormap.restype = ctypes.c_ulong
_libX11.XDefaultColormap.argtypes = (ctypes.c_void_p, ctypes.c_int)

# int XAllocNamedColor(Display *display, Colormap cmap, char *color_name,
#                  XColor *screen_def_return, XColor *exact_def_return)
_libX11.XAllocNamedColor.restype = ctypes.c_int
_libX11.XAllocNamedColor.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
    ctypes.c_char_p, ctypes.POINTER(XColor), ctypes.POINTER(XColor))

# Pixmap XCreateBitmapFromData(Display *display, Drawable d, char *data,
#                              uint width, uint height)
_libX11.XCreateBitmapFromData.restype = ctypes.c_ulong
_libX11.XCreateBitmapFromData.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
    ctypes.c_char_p, ctypes.c_uint, ctypes.c_uint)

# void XFreePixmap(Display *display, Pixmap p)
_libX11.XFreePixmap.argtypes = (ctypes.c_void_p, ctypes.c_ulong)

# void XFreeColors(Display *display, Colormap colormap, ulong pixels[],
#                  int npixels, ulong planes))
_libX11.XFreeColors.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
    ctypes.POINTER(ctypes.c_ulong), ctypes.c_int, ctypes.c_ulong)

# XResizeWindow(Display *display, Window w, uint width, uint height)
_libX11.XResizeWindow.argtypes = (ctypes.c_void_p, ctypes.c_ulong,
                                  ctypes.c_uint, ctypes.c_uint)


# copy functions from the library
module = sys.modules[__name__]
for name in dir(_libX11):
    if name.startswith('X'):
        setattr(module, name, getattr(_libX11, name))


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


import re
from deejayd import mediafilters
from deejayd.database.querybuilders import EditRecordQuery


class _DBObject(object):

    def __init__(self):
        self.id = None

    def save(self, db):
        raise NotImplementedError


class NoneFilter(_DBObject):

    CLASS_TABLE = 'filters'
    PRIMARY_KEY = 'filter_id'
    TYPE = 'none'

    def __init__(self, none_filter = None):
        super(NoneFilter, self).__init__()
        self.id = None

    def restrict(self, query):
        pass

    def save(self, db):
        cursor = db.connection.cursor()

        query = EditRecordQuery(self.CLASS_TABLE)
        query.add_value('type', self.TYPE)
        if self.id:
            query.set_update_id(self.PRIMARY_KEY, self.id)
        cursor.execute(query.to_sql(), query.get_args())
        if not self.id:
            self.id = db.connection.get_last_insert_id(cursor)

        cursor.close()

        return self.id


class _BasicFilter(mediafilters.BasicFilter, NoneFilter):

    TABLE = 'filters_basicfilters'
    TYPE = 'basic'

    def __init__(self, basic_filter):
        NoneFilter.__init__(self)
        mediafilters.BasicFilter.__init__(self, basic_filter.tag,
                                                basic_filter.pattern)

    def restrict(self, query):
        query.join_on_tag(self.tag)
        where_str, arg = self._match_tag()
        query.append_where(where_str, (arg,))

    def _match_tag(self, match_value):
        raise NotImplementedError

    def save(self, db):
        if not self.id:
            new = True
        else:
            new = False

        super(_BasicFilter, self).save(db)

        cursor = db.connection.cursor()

        query = EditRecordQuery(self.TABLE)
        if new:
            query.add_value(self.PRIMARY_KEY, self.id)
        else:
            query.set_update_id(self.PRIMARY_KEY, self.id)
        query.add_value('tag', self.tag)
        query.add_value('operator', self.get_name())
        query.add_value('pattern', self.pattern)
        cursor.execute(query.to_sql(), query.get_args())

        cursor.close()

        return self.id


class Equals(mediafilters.Equals, _BasicFilter):

    def _match_tag(self):
        return "(%s.value = " % (self.tag,) + "%s)", self.pattern


class NotEquals(mediafilters.NotEquals, _BasicFilter):

    def _match_tag(self):
        return "(%s.value != " % (self.tag,) + "%s)", self.pattern


class Contains(mediafilters.Contains, _BasicFilter):

    def _match_tag(self):
        return "(%s.value LIKE " % (self.tag,) + "%s)", "%%"+self.pattern+"%%"


class NotContains(mediafilters.NotContains, _BasicFilter):

    def _match_tag(self):
        return "(%s.value NOT LIKE " % (self.tag,)+"%s)", "%%"+self.pattern+"%%"


class Higher(mediafilters.Higher, _BasicFilter):

    def _match_tag(self):
        return "(%s.value >= " % (self.tag,)+"%s)", self.pattern


class Lower(mediafilters.Lower, _BasicFilter):

    def _match_tag(self):
        return "(%s.value <= " % (self.tag,)+"%s)", self.pattern


class Regexi(mediafilters.Regexi, _BasicFilter):
    # FIXME : to implement some day
    pass


class _ComplexFilter(mediafilters.ComplexFilter, NoneFilter):

    TABLE = 'filters_complexfilters'
    TYPE = 'complex'

    def __init__(self, complex_filter):
        self.sqlfilterlist = map(SQLizer().translate, complex_filter.filterlist)
        arglist = [self] + self.sqlfilterlist
        mediafilters.ComplexFilter.__init__(*arglist)

    def restrict(self, query):
        if self.filterlist:
            where_str, args = self._build_wheres(query)
            query.append_where(where_str, args)

    def _build_wheres(self, query):
        if not self.filterlist: return "(1)", []
        wheres, wheres_args = [], []
        for filter in self.filterlist:
            if filter.type == "basic":
                query.join_on_tag(filter.tag)
                where_query, arg = filter._match_tag()
                wheres.append(where_query)
                wheres_args.append(arg)
            else: # complex filter
                where_query, args = filter._build_wheres(query)
                wheres.append(where_query)
                wheres_args.extend(args)
        return "(%s)" % (self.combinator.join(wheres),), wheres_args

    def save(self, db):
        if not self.id:
            new = True
        else:
            new = False
        super(_ComplexFilter, self).save(db)

        cursor = db.connection.cursor()

        query = EditRecordQuery(_ComplexFilter.TABLE)
        if new:
            query.add_value(self.PRIMARY_KEY, self.id)
        else:
            query.set_update_id(self.PRIMARY_KEY, self.id)
        query.add_value('combinator', self.get_name())
        cursor.execute(query.to_sql(), query.get_args())

        for subfilter in self.sqlfilterlist:
            subfilter_id = subfilter.save(db)
            query = EditRecordQuery('filters_complexfilters_subfilters')
            query.add_value('complexfilter_id', self.id)
            query.add_value('filter_id', subfilter_id)
            cursor.execute(query.to_sql(), query.get_args())

        cursor.close()

        return self.id


class And(mediafilters.And, _ComplexFilter): combinator = ' AND '
class Or(mediafilters.Or, _ComplexFilter):   combinator = ' OR '


class SQLizer(object):

    translations = {
                     mediafilters.Equals :    Equals,
                     mediafilters.NotEquals : NotEquals,
                     mediafilters.Contains :  Contains,
                     mediafilters.NotContains :  NotContains,
                     mediafilters.Regexi :    Regexi,
                     mediafilters.Higher :    Higher,
                     mediafilters.Lower :    Lower,
                     mediafilters.And :       And,
                     mediafilters.Or :        Or,
                   }

    def translate(self, object):
        if object == None:
            return NoneFilter(object)
        object_class = SQLizer.translations[object.__class__]
        return object_class(object)

# vim: ts=4 sw=4 expandtab



# -*- coding: utf-8 -*-
# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

songlibrary = [
{"media_id":1, 'filename': 'rock/Dolly - Dolly/01-dolly-je_n_veux_pas_rester_sage.ogg',
 'artist': 'Dolly', 'album': 'Dolly', 'title': "Je n'veux pas rester sage"},
{"media_id":2, 'filename': 'rap/MC Solar/LTDLQ/04-mc_solaar-onzième_commandement.ogg',
 'artist': 'MC Solar', 'album': 'Le tour de la question CD1',
 'title': "Onzième commandement", 'track': '04/20'},
{"media_id":3, 'filename': 'dance/singles/Ultra Nate - Free.mp3',
 'artist': 'Ultra Nate', 'title': "Free"}
]

sample_genres = ('Country', 'Disco', 'Rock', 'Trance', 'R&B', 'Big beat', )


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.


from deejayd.component import SignalingComponent
from deejayd.player._base import PlayerError
from deejayd.player.xine import DvdParser

class DvdError(Exception): pass

class DvdSource(SignalingComponent):
    name = "dvd"

    def __init__(self, db, config):
        super(DvdSource, self).__init__()
        try: self.parser = DvdParser()
        except PlayerError, ex: # dvd parser does not work
            raise DvdError(ex)

        self.db = db
        try: self.current_id = int(self.db.get_state("dvdid"))
        except TypeError: # init dvdid
            self.current_id = 1

        self.dvd_info = None
        self.selected_track = None

    def get_content(self):
        if not self.dvd_info:
            return {'title': "DVD NOT LOADED", 'longest_track': "0", \
                    'track': []}
        return self.dvd_info

    def load(self):
        # Reinit var and update id
        self.dvd_info = None
        self.selected_track = None
        self.current_id +=1

        try: self.dvd_info = self.parser.get_dvd_info()
        except PlayerError, err:
            raise DvdError("Unable to load the dvd : %s " % err)
        # select the default track of the dvd
        self.select_track()
        self.dispatch_signame('dvd.update')

    def select_track(self,nb = None,alang_idx = None, slang_idx = None):
        if not self.dvd_info: return

        if not nb: nb = self.dvd_info['longest_track']
        for track in self.dvd_info['track']:
            if int(track['ix']) == int(nb):
                self.selected_track = track
                self.selected_track["selected_chapter"] = -1
                break

    def select_chapter(self,track = None, chapter = 1, alang = None, \
                        slang = None):
        self.selected_track = self.select_track(track,alang,slang)
        self.selected_track["selected_chapter"] = chapter

    def get_current(self):
        if not self.dvd_info or not self.selected_track: return None

        # Construct item info for player
        id = str(self.selected_track["ix"])
        pos = int(self.selected_track["ix"])
        if self.selected_track["selected_chapter"] != -1:
            id += ".%d" % self.selected_track["selected_chapter"]
            pos = self.selected_track["selected_chapter"]
        uri = "dvd://%d" % self.selected_track["ix"]
        return {"title": self.dvd_info["title"], "type": "video", \
                "uri": uri, "chapter":self.selected_track["selected_chapter"],\
                "length": self.selected_track["length"],\
                "id": id, "pos": pos,\
                "audio": self.selected_track["audio"],\
                "subtitle": self.selected_track["subp"]}

    def go_to(self,id,type = "track"):
        if type in ("track", "id"): self.select_track(id)
        elif type == "chapter": self.select_chapter(None,id)
        elif type == "dvd_id":
            ids = id.split('.')
            self.select_track(int(ids[0]))
            if len(ids) > 1 and self.selected_track:
                self.selected_track["selected_chapter"] = int(ids[1])

        return self.get_current()

    def next(self, explicit = True):
        if not self.dvd_info or not self.selected_track: return None

        if self.selected_track["selected_chapter"] != -1:
            # go to next chapter
            self.selected_track["selected_chapter"] += 1
        else:
            # go to next title
            current_title = self.selected_track["ix"]
            self.select_track(current_title+1)

        return self.get_current()


    def previous(self):
        if not self.dvd_info or not self.selected_track: return None

        if self.selected_track["selected_chapter"] != -1:
            # go to previous chapter
            self.selected_track["selected_chapter"] -= 1
        else:
            # go to previous title
            current_title = self.selected_track["ix"]
            self.select_track(current_title - 1)

        return self.get_current()

    def get_status(self):
        length = 0
        if self.dvd_info: length = len(self.dvd_info)
        status = [
            (self.name, self.current_id),
            (self.name+"length",length)
            ]
        return status

    def close(self):
        states = [(str(self.current_id),"dvdid")]
        self.db.set_state(states)
        self.parser.close()

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
Tools to create a test server.
"""
import os, signal, os.path, sys, subprocess

logfiles = ['/tmp/testdeejayd.log', '/tmp/testdeejayd-webui.log']
for logfile in logfiles:
    if os.path.isfile(logfile):
        os.unlink(logfile)


class TestServer:
    """Implements a server ready for testing."""

    def __init__(self, conf_file):
        self.conf_file = conf_file
        self.serverExecRelPath = 'scripts/testserver'
        self.srcpath = self.findSrcPath()

    def findSrcPath(self):
        # Get the server executable path, assuming it is names
        # scripts/testserver in a subdirectory of $PYTHONPATH
        absPath = ''
        notFound = True
        sysPathIterator = iter(sys.path)
        while notFound:
            absPath = sysPathIterator.next()
            serverScriptPath = os.path.join(absPath, self.serverExecRelPath)
            if os.path.exists(serverScriptPath):
                notFound = False

        if notFound:
            raise Exception('Cannot find server executable')

        return os.path.abspath(absPath)

    def start(self):
        serverExec = os.path.join(self.srcpath, self.serverExecRelPath)

        if not os.access(serverExec, os.X_OK):
            sys.exit("The test server executable '%s' is not executable."\
                     % serverExec)

        args = [serverExec, self.conf_file]
        env = {'PYTHONPATH': self.srcpath, "PATH": os.getenv('PATH'),\
                'LANG': os.getenv('LANG')}
        self.__serverProcess = subprocess.Popen(args = args,
                                                env = env,
                                                stderr = subprocess.PIPE,
                                                stdout = sys.stdout.fileno(),
                                                close_fds = True)

        firstLine = self.__serverProcess.stderr.readline()
        if not firstLine == 'ready\n':
            # Should not occur
            print firstLine
            self.stop()
            raise Exception('Reactor does not seem to be ready')

    def stop(self):
        # Send stop signal to reactor
        os.kill(self.__serverProcess.pid, signal.SIGINT)

        # Wait for the process to finish
        self.__serverProcess.wait()


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os

from testdeejayd import TestCaseWithAudioAndVideoData
from testdeejayd.coreinterface import InterfaceTests, InterfaceSubscribeTests

from deejayd.core import DeejayDaemonCore
from deejayd.ui.config import DeejaydConfig


class TestCore(TestCaseWithAudioAndVideoData, InterfaceTests,
                                              InterfaceSubscribeTests):
    """Test the deejayd daemon core."""

    def setUp(self):
        TestCaseWithAudioAndVideoData.setUp(self)

        config = DeejaydConfig()
        self.dbfilename = '/tmp/testdeejayddb-' +\
                          self.testdata.getRandomString() + '.db'

        config.set('general', 'activated_modes',\
                'playlist,panel,webradio,video,dvd')
        config.set('database', 'db_type', 'sqlite')
        config.set('database', 'db_name', self.dbfilename)

        config.set('mediadb','music_directory',self.test_audiodata.getRootDir())
        config.set('mediadb','video_directory',self.test_videodata.getRootDir())

        # player option
        config.set('general','media_backend',"xine")
        config.set('general','video_support',"yes")
        config.set('xine','audio_output',"none")
        config.set('xine','video_output',"none")

        self.deejayd = DeejayDaemonCore(config)
        self.deejayd.audio_library._update()
        self.deejayd.video_library._update()

    def tearDown(self):
        self.deejayd.close()
        os.unlink(self.dbfilename)
        TestCaseWithAudioAndVideoData.tearDown(self)

    def test_objanswer_mechanism(self):
        """Test the objanswer mechanism to disable DeejaydAnswer objects in returns parameters."""
        known_mode = 'playlist'

        # objanswer mechanism on (default)
        ans = self.deejayd.set_mode(known_mode)
        self.failUnless(ans.get_contents())
        ans = self.deejayd.get_status()
        self.assertEqual(ans.get_contents()['mode'], known_mode)

        # objanswer mechanism off
        ans = self.deejayd.set_mode(known_mode, objanswer=False)
        self.failUnless(ans == None)
        ans = self.deejayd.get_status(objanswer=False)
        self.assertEqual(ans['mode'], known_mode)


# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd import __version__ as VERSION

def build(config):
    return """
<?xml-stylesheet href="chrome://global/skin/" type="text/css"?>

<window
    id="deejayd-webui-window"
    title="Deejayd webui"
    xmlns:html="http://www.w3.org/1999/xhtml"
    xmlns="http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul">

    <script>
        window.onload = function(e) {
            var element = document.createElement("deejaydExtension");
            element.setAttribute("refresh", '%(refresh)s');
            element.setAttribute("version", '%(version)s');
            document.documentElement.appendChild(element);

            var evt = document.createEvent("Events");
            evt.initEvent("deejaydEvent", true, false);
            element.dispatchEvent(evt);
            };
    </script>
    <vbox>
        <vbox id="deejayd-webui_install">
            <groupbox>
                <vbox style="padding: 10px">
                    <description>
                        %(install)s
                    </description>
                </vbox>
                <vbox align="center">
                    <html:a href="../static/deejayd-webui.xpi">
                        %(clickHere)s
                    </html:a>
                </vbox>
            </groupbox>
        </vbox>
        <vbox id="deejayd-webui_upgrade" style="display:none">
            <groupbox>
                <vbox style="padding: 10px">
                    <description>
                        %(upgrade)s
                    </description>
                </vbox>
                <vbox align="center">
                    <html:a href="../static/deejayd-webui.xpi">
                        %(clickHere)s
                    </html:a>
                </vbox>
            </groupbox>
        </vbox>
        <description style="display:none;color:#f00"
            id="deejayd-webui_error" value="%(error)s"/>
    </vbox>
</window>
""" % {
        "refresh": config.get('webui','refresh'),
        "version": VERSION,
        "install": _("You need to install a firefox extension in order to use the deejayd-webui XUL client. Please note that if you run a flavour of GNU/Linux, it should be available from your package manager.").\
                encode("utf-8"),
        "upgrade": _("You need to upgrade the firefox extension.").\
                encode("utf-8"),
        "clickHere": _("Install the deejayd-webui extension").encode("utf-8"),
        "error": _("ERROR : Host is not allowed to use the firefox extension.").\
                encode("utf-8"),
    }

# vim: ts=4 sw=4 expandtab



# -*- coding: utf-8 -*-
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,

"""
This module generates the test data.
"""
import os, sys, shutil, urllib, random, time, string

from deejayd.mediafilters import *

from testdeejayd.data import sample_genres

DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')


class TestDataError: pass

class TestData(object):

    def __init__(self):
        # Grab some sample data
        import testdeejayd.data
        self.sampleLibrary = testdeejayd.data.songlibrary

    def getRandomString(self, length = 5, charset = string.letters,\
                        special = False):
        random.seed(time.time())
        rs = ''.join(random.sample(charset, length-1))
        #if special:
        #    rs += ''.join(random.sample(['é','è',"'","`","?","-","ç","à"], 2))
        return rs

    def getRandomInt(self, maxValue = 10, minValue = 1):
        random.seed(time.time())
        return random.randint(minValue, maxValue)

    def getRandomElement(self, list):
        random.seed(time.time())
        return random.sample(list, 1).pop()

    def getBadCaracter(self):
        return "\xe0"

    def getRandomGenre(self):
        return self.getRandomElement(sample_genres)

    def getSampleParmDict(self, howMuch = 2):
        parmDict = {}
        for i in range(howMuch):
            parmDict['parmName' + str(i)] = 'parmValue' + str(i)
        return parmDict

    def get_sample_filter(self):
        filter = And(Contains('artist', 'Britney'),
                     Or(Equals('genre', 'Classical'),
                        Equals('genre', 'Disco')
                     )
                    )
        return filter


class TestSong(TestData):
    supportedTag = ("tracknumber","title","genre","artist","album","date")

    def __init__(self):
        self.name = self.getRandomString(special = True) + self.ext
        self.tags = {}

    def build(self,path):
        filename = os.path.join(path, self.name)
        shutil.copy(self.testFile, filename)
        self.tags["filename"] = filename
        self.tags["uri"] = "file://%s" % urllib.quote(filename)

        self.setRandomTag()

    def remove(self):
        os.unlink(self.tags["filename"])
        self.tags = None

    def rename(self):
        newName = self.getRandomString(special = True) + self.ext
        newPath = os.path.join(os.path.dirname(self.tags['filename']),\
            newName)
        os.rename(self.tags['filename'],newPath)
        self.tags["filename"] = newPath
        self.name = newName

    def get_random_tag_value(self, tagname):
        if tagname == "date":
            value = str(self.getRandomInt(2010,1971))
        elif tagname == "tracknumber":
            value = str(self.getRandomInt(15))
        elif tagname == "genre":
            value = self.getRandomGenre()
        else:
            value = self.getRandomString(special=False)
        return value

    def setRandomTag(self):
        tagInfo = self._getTagObject()
        for tag in self.__class__.supportedTag:
            value = self.get_random_tag_value(tag)
            tagInfo[tag] = unicode(value)
            self.tags[tag] = value
        tagInfo.save()


class TestVideo(TestSong):

    def __init__(self):
        self.testFile,self.ext = os.path.join(DATA_DIR,"video_test.avi"),".avi"
        super(TestVideo, self).__init__()
        # FIXME Shoudn't videowidth and videoheight be of type int?
        self.tags = {"length": 2, "videowidth": '640', "videoheight": '480',\
                     "external_subtitle": ""}

    def __getitem__(self,key):
        try: return self.tags[key]
        except KeyError:
            return None

    def set_subtitle(self, sub):
        if sub != "":
            self.tags["external_subtitle"] = "file://"+sub
        else:
            self.tags["external_subtitle"] = ""

    def setRandomTag(self):pass


class TestMP3Song(TestSong):

    def __init__(self):
        self.testFile,self.ext = os.path.join(DATA_DIR, "mp3_test.mp3"), ".mp3"
        #self.name = self.getRandomString() + self.getBadCaracter() + self.ext
        self.name = self.getRandomString() + self.ext
        self.tags = {}

    def __getitem__(self,key):
        return key in self.tags and self.tags[key] or None

    def _getTagObject(self):
        from mutagen.easyid3 import EasyID3
        return EasyID3(self.tags["filename"])


class TestMP4Song(TestSong):
    __translate = {
        "\xa9nam": "title",
        "\xa9alb": "album",
        "\xa9ART": "artist",
        "\xa9day": "date",
        "\xa9gen": "genre",
        }
    __tupletranslate = {
        "trkn": "tracknumber",
        }

    def __init__(self):
        self.testFile,self.ext = os.path.join(DATA_DIR, "mp4_test.mp4"), ".mp4"
        super(TestMP4Song, self).__init__()

    def __getitem__(self,key):
        return key in self.tags and self.tags[key] or None

    def setRandomTag(self):
        from mutagen.mp4 import MP4
        tag_info = MP4(self.tags["filename"])
        for tag, name in self.__translate.iteritems():
            value = self.get_random_tag_value(name)
            tag_info[tag] = unicode(value)
            self.tags[name] = value

        for tag, name in self.__tupletranslate.iteritems():
            cur = self.getRandomInt(15)
            value = (cur, 15)
            tag_info[tag] = [value]
            self.tags[name] = "%d/15" % cur

        tag_info.save()

class TestOggSong(TestSong):

    def __init__(self):
        self.testFile,self.ext = os.path.join(DATA_DIR, "ogg_test.ogg"), ".ogg"
        super(TestOggSong, self).__init__()

    def __getitem__(self,key):
        return key in self.tags and self.tags[key] or None

    def _getTagObject(self):
        from mutagen.oggvorbis import OggVorbis
        return OggVorbis(self.tags["filename"])


class TestFlacSong(TestSong):

    def __init__(self):
        self.testFile,self.ext = os.path.join(DATA_DIR,"flac_test.flac"),".flac"
        super(TestFlacSong, self).__init__()

    def __getitem__(self,key):
        return key in self.tags and self.tags[key] or None

    def _getTagObject(self):
        from mutagen.flac import FLAC
        return FLAC(self.tags["filename"])


class _TestDir(TestData):

    def __init__(self):
        self.build = False
        self.name = self.getRandomString(special = True)
        self.root = None
        self.items = []

    def addItem(self, item):
        if self.build: item.build(self.dirPath)
        self.items.append(item)

    def buildContent(self, destDir):
        self.dirPath = os.path.join(destDir,self.name)
        os.mkdir(self.dirPath)
        time.sleep(0.1)
        for item in self.items:
            item.build(self.dirPath)
        self.build = True
        self.root = destDir

    def rename(self):
        newName = self.getRandomString(special = True)
        os.rename(os.path.join(self.root,self.name),\
                  os.path.join(self.root,newName))
        self.name = newName

    def remove(self):
        shutil.rmtree(self.dirPath)
        self.build = False


class TestAudioDir(_TestDir):

    def __init__(self):
        super(TestAudioDir, self).__init__()
        self.cover = None

    def buildContent(self, destDir, with_cover = False):
        super(TestAudioDir, self).buildContent(destDir)
        if with_cover: self.add_cover()

    def add_cover(self):
        if not self.cover:
            cover_path = os.path.join(DATA_DIR, "cover.jpg")
            self.cover = os.path.join(self.dirPath, "cover.jpg")
            shutil.copy(cover_path, self.cover)

    def remove_cover(self):
        if self.cover:
            os.unlink(self.cover)
            self.cover = None


class TestVideoDir(_TestDir):

    def __init__(self):
        super(TestVideoDir, self).__init__()
        self.has_sub = False

    def buildContent(self, destDir, with_subtitle = False):
        super(TestVideoDir, self).buildContent(destDir)
        if with_subtitle:
            self.add_subtitle()

    def add_subtitle(self):
        sub_path = os.path.join(DATA_DIR, "sub.srt")
        for item in self.items:
            if not item["external_subtitle"]:
                sub_name = os.path.basename(item["filename"])
                (sub_name, ext) = os.path.splitext(sub_name)
                dest = os.path.join(self.dirPath, sub_name + ".srt")
                shutil.copy(sub_path, dest)
                item.set_subtitle(dest)
        self.has_sub = True

    def remove_subtitle(self):
        for item in self.items:
            if item["external_subtitle"]:
                os.unlink(item["external_subtitle"].replace("file://", ""))
                item.set_subtitle("")
        self.has_sub = False


class TestProvidedMusicCollection(TestData):

    def __init__(self, musicDir):
        self.datadir = os.path.normpath(musicDir)

        self.songPaths = []
        for root, dir, files in os.walk(self.datadir):
            for file in files:
                (name,ext) = os.path.splitext(file)
                if ext.lower() in ('.mp3','.ogg','.mp4','.flac'):
                    self.songPaths.append(self.stripRoot(os.path.join(root,
                                                                        file)))

    def getRootDir(self):
        return self.datadir

    def get_song_paths(self):
        return self.songPaths

    def stripRoot(self, path):
        """Strips the root directory path turning the argument into a
        path relative to the music root directory."""
        abs_path = os.path.abspath(path)
        rel_path = os.path.normpath(abs_path[len(self.getRootDir()):])

        if rel_path != '.': rel_path = rel_path.strip("/")
        else: rel_path = ''

        return rel_path

    def getRandomSongPaths(self, howMuch = 1):
        """Returns the path of a random song in provided music"""
        random.seed(time.time())
        return random.sample(self.get_song_paths(), howMuch)


class _TestMediaCollection(TestProvidedMusicCollection):

    def __init__(self):
        self.dir_struct_written = False
        self.clean_library = True
        self.dirs = {}
        self.dirlinks = {}
        self.medias = {}
        self.supported_files_class = ()

    def cleanLibraryDirectoryTree(self):
        if self.dir_struct_written and self.clean_library:
            shutil.rmtree(self.datadir)
            for dirlink in self.dirlinks.values():
                shutil.rmtree(dirlink.datadir)

    def buildLibraryDirectoryTree(self, destDir = "/tmp"):
        # create test data directory in random subdirectory of destDir
        self.datadir = os.path.join(os.path.normpath(destDir),
                                   'testdeejayd-media' + self.getRandomString())
        if not os.path.exists(self.datadir):
            os.mkdir(self.datadir)
        else:
            sys.exit(\
     'Test data temporary directory exists, I do not want to mess your stuff.')

        # Add songs in the root directory
        for media_class in self.__class__.supported_files_class:
            media = media_class()
            media.build(self.datadir)
            self.medias[media.name] = media

        # Create several directories
        for i in range(self.getRandomInt(10,5)):
            self.addDir()

        # Add a subdirectory
        self.addSubdir()

        self.dir_struct_written = True

    def get_song_paths(self):
        song_paths = self.medias.keys()
        for dirlink in self.dirlinks.values():
            for song_path in dirlink.get_song_paths():
                song_paths.append(song_path)
        return song_paths

    def addMedia(self):
        dir = self.getRandomElement(self.dirs.values())

        media_class=self.getRandomElement(self.__class__.supported_files_class)
        media = media_class()
        dir.addItem(media)
        self.medias[os.path.join(dir.name, media.name)] = media

    def renameMedia(self):
        mediaKey = self.getRandomElement(self.medias.keys())
        media = self.medias[mediaKey]
        del self.medias[mediaKey]

        media.rename()
        new_path = os.path.join(os.path.dirname(mediaKey), media.name)
        self.medias[new_path] = media

    def removeMedia(self):
        mediaKeys = self.getRandomElement(self.medias.keys())
        self.medias[mediaKeys].remove()
        del self.medias[mediaKeys]

    def addDir(self):
        dir = self.dir_class()
        for media_class in self.__class__.supported_files_class:
            media = media_class()
            self.medias[os.path.join(dir.name, media.name)] = media
            dir.addItem(media)
        dir.buildContent(self.datadir, random.choice((True, False)))

        self.dirs[dir.name] = dir

    def addSubdir(self):
        dir = self.getRandomElement(self.dirs.values())
        subdir = self.dir_class()
        for media_class in self.__class__.supported_files_class:
            media = media_class()
            media_path = os.path.join(dir.name, subdir.name, media.name)
            self.medias[media_path] = media
            subdir.addItem(media)

        subdir_path = os.path.join(self.datadir, dir.name)
        subdir.buildContent(subdir_path)
        self.dirs[subdir_path] = dir

    def addSubSubdir(self):
        dir = self.getRandomElement(self.dirs.values())
        subdir, subsubdir = self.dir_class(), self.dir_class()
        for media_class in self.__class__.supported_files_class:
            media = media_class()
            media_path = os.path.join(dir.name, subdir.name, subsubdir.name,\
                    media.name)
            self.medias[media_path] = media
            subsubdir.addItem(media)

        subdir_path = os.path.join(self.datadir, dir.name)
        subdir.buildContent(subdir_path)
        self.dirs[subdir_path] = dir

        subsubdir_path = os.path.join(self.datadir, dir.name, subdir.name)
        subsubdir.buildContent(subsubdir_path)
        self.dirs[subsubdir_path] = subdir

    def renameDir(self):
        dir = self.getRandomElement(self.dirs.values())
        del self.dirs[dir.name]

        dir.rename()
        self.dirs[dir.name] = dir

    def removeDir(self):
        dir = self.getRandomElement(self.dirs.values())
        dir.remove()
        del self.dirs[dir.name]

    def addDirLink(self):
        where = self.getRandomElement(self.dirs.values())
        dirlink = self.__class__()
        linkname = self.getRandomString()
        linkpath = os.path.join(self.datadir, where.name, linkname)
        dirlink.buildLibraryDirectoryTree()
        self.dirlinks[linkpath] = dirlink

        os.symlink(dirlink.datadir, linkpath)

    def moveDirLink(self):
        dirlinkpath = self.getRandomElement(self.dirlinks.keys())
        linkname = os.path.basename(dirlinkpath)

        new_location = self.getRandomElement([d for d in self.dirs.keys()\
                                              if d != linkname])
        new_location = os.path.join(self.getRootDir(), new_location, linkname)
        os.rename(dirlinkpath, new_location)

        dirlink = self.dirlinks[dirlinkpath]
        del self.dirlinks[dirlinkpath]
        self.dirlinks[new_location] = dirlink

    def removeDirLink(self):
        dirlinkpath, dirlink = self.getRandomElement(self.dirlinks.items())
        os.unlink(dirlinkpath)
        dirlink.cleanLibraryDirectoryTree()
        del self.dirlinks[dirlinkpath]


class TestAudioCollection( _TestMediaCollection):
    dir_class = TestAudioDir
    supported_files_class = (TestOggSong,TestMP3Song,TestMP4Song,TestFlacSong)

    def remove_cover(self):
        for dir in self.dirs.values():
            if dir.cover:
                dir.remove_cover()
                return

    def add_cover(self):
        for dir in self.dirs.values():
            if not dir.cover:
                dir.add_cover()
                return

    def changeMediaTags(self):
        media = self.getRandomElement(self.medias.values())
        media.setRandomTag()


class TestVideoCollection( _TestMediaCollection):
    dir_class = TestVideoDir
    supported_files_class = (TestVideo,)

    def remove_subtitle(self):
        for dir in self.dirs.values():
            if dir.has_sub:
                dir.remove_subtitle()
                return

    def add_subtitle(self):
        for dir in self.dirs.values():
            if not dir.has_sub:
                dir.add_subtitle()
                return

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from deejayd.sources._base import _BaseAudioLibSource

class PlaylistSource(_BaseAudioLibSource):
    SUBSCRIPTIONS = {
            "mediadb.mupdate": "cb_library_changes",
            }
    base_medialist = "__djcurrent__"
    name = "playlist"
    source_signal = 'player.plupdate'

    def shuffle(self):
        self._media_list.shuffle(self._current)
        if self._current: self._current["pos"] = 0
        self.dispatch_signame(self.__class__.source_signal)

    def save(self, playlist_name):
        id = self.db.set_static_medialist(playlist_name, self._media_list.get())
        self.dispatch_signame('playlist.listupdate')
        return {"playlist_id": id}

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
# -*- coding: utf-8 -*-

import os, sys, threading, traceback, base64, locale, hashlib
from twisted.internet import threads, reactor
import kaa.metadata


from deejayd.interfaces import DeejaydError
from deejayd.component import SignalingComponent
from deejayd.mediadb import formats
from deejayd.utils import quote_uri, str_encode
from deejayd import database, mediafilters
from deejayd.ui import log

class NotFoundException(DeejaydError):pass
class NotSupportedFormat(DeejaydError):pass


##########################################################################
##########################################################################
def inotify_action(func):
    def inotify_action_func(*__args, **__kw):
        self = __args[0]
        try:
            name = self._encode(__args[1])
            path = self._encode(__args[2])
        except UnicodeError:
            return

        return func(*__args, **__kw)

    return inotify_action_func


class _Library(SignalingComponent):
    common_attr = ("filename","uri","type","title","length")
    persistent_attr = ("rating","skipcount","playcount","lastplayed")
    type = None

    def __init__(self, db_connection, player, path, fs_charset="utf-8"):
        SignalingComponent.__init__(self)

        # init Parms
        self.media_attr = []
        for i in self.__class__.common_attr: self.media_attr.append(i)
        for j in self.__class__.custom_attr: self.media_attr.append(j)
        for k in self.__class__.persistent_attr: self.media_attr.append(k)
        self._fs_charset = fs_charset
        self._update_id = 0
        self._update_end = True
        self._update_error = None

        self._changes_cb = {}
        self._changes_cb_id = 0

        self._path = os.path.abspath(path)
        # test library path
        if not os.path.isdir(self._path):
            msg = _("Unable to find directory %s") % self._encode(self._path)
            raise NotFoundException(msg)

        # Connection to the database
        self.db_con = db_connection

        # init a mutex
        self.mutex = threading.Lock()

        # build supported extension list
        self.ext_dict = formats.get_extensions(player, self.type)

        self.watcher = None

    def _encode(self, data):
        return str_encode(data, self._fs_charset)

    def _build_supported_extension(self, player):
        raise NotImplementedError

    def set_file_info(self, file_id, key, value, allow_create = False):
        ans = self.db_con.set_media_infos(file_id, {key: value}, allow_create)
        if not ans:
            raise NotFoundException
        self.dispatch_signame('mediadb.mupdate',\
                attrs = {"type": "update", "id": file_id})
        self.db_con.connection.commit()

    def get_dir_content(self, dir):
        dir = os.path.join(self._path, dir).rstrip("/")
        files_rsp = self.db_con.get_dir_content(dir,\
            infos = self.media_attr, type = self.type)
        dirs_rsp = self.db_con.get_dir_list(dir, self.type)
        if len(files_rsp) == 0 and len(dirs_rsp) == 0 and dir != self._path:
            # nothing found for this directory
            raise NotFoundException

        dirs = []
        for dir_id, dir_path in dirs_rsp:
            root, d = os.path.split(dir_path.rstrip("/"))
            if d != "" and root == self._encode(dir):
                dirs.append(d)
        return {'files': files_rsp, 'dirs': dirs}

    def get_dir_files(self,dir):
        dir = os.path.join(self._path, dir).rstrip("/")
        files_rsp = self.db_con.get_dir_content(dir,\
            infos = self.media_attr, type = self.type)
        if len(files_rsp) == 0 and dir != self._path: raise NotFoundException
        return files_rsp

    def get_all_files(self,dir):
        dir = os.path.join(self._path, dir).rstrip("/")
        files_rsp = self.db_con.get_alldir_files(dir,\
            infos = self.media_attr, type = self.type)
        if len(files_rsp) == 0 and dir != self._path: raise NotFoundException
        return files_rsp

    def get_file(self,file):
        file = os.path.join(self._path, file)
        d, f = os.path.split(file)
        files_rsp = self.db_con.get_file(d, f,\
            infos = self.media_attr, type = self.type)
        if len(files_rsp) == 0:
            # this file is not found
            raise NotFoundException
        return files_rsp

    def get_file_withids(self,file_ids):
        files_rsp = self.db_con.get_file_withids(file_ids,\
            infos = self.media_attr, type = self.type)
        if len(files_rsp) != len(file_ids):
            raise NotFoundException
        return files_rsp

    def search(self, filter, ords = [], limit = None):
        ft = mediafilters.And()
        ft.combine(mediafilters.Equals("type", self.__class__.search_type))
        if filter is not None:
            ft.combine(filter)

        return self.db_con.search(ft, infos = self.media_attr, orders=ords,\
                limit = limit)

    def get_root_path(self):
        return self._path

    def get_root_paths(self):
        root_paths = [self.get_root_path()]
        for id, dirlink_record in self.db_con.get_all_dirlinks('', self.type):
            dirlink = os.path.join(self.get_root_path(), dirlink_record)
            root_paths.append(dirlink)
        return root_paths

    def get_status(self):
        status = []
        if not self._update_end:
            status.append((self.type+"_updating_db",self._update_id))
        if self._update_error:
            status.append((self.type+"_updating_error",self._update_error))
            self._update_error = None

        return status

    def close(self):
        pass

    #
    # Update process
    #
    def update(self, force = False, sync = False):
        if self._update_end:
            self._update_id += 1
            if sync: # synchrone update
                self._update(force)
                self._update_end = True
            else: # asynchrone update
                self.defered = threads.deferToThread(self._update, force)
                self.defered.pause()

                # Add callback functions
                succ = lambda *x: self.end_update()
                self.defered.addCallback(succ)

                # Add errback functions
                def error_handler(failure,db_class):
                    # Log the exception to debug pb later
                    failure.printTraceback()
                    db_class.end_update(False)
                    return False
                self.defered.addErrback(error_handler,self)

                self.defered.unpause()
            self.dispatch_signame(self.update_signal_name)
            return self._update_id

        return 0

    def is_in_root(self, path, root=None):
        """Checks if a directory is physically in the supplied root (the library root by default)."""
        if not root:
            root = self.get_root_path()
        real_root = os.path.realpath(root)
        real_path = os.path.realpath(path)

        head = real_path
        old_head = None
        while head != old_head:
            if head == real_root:
                return True
            old_head = head
            head, tail = os.path.split(head)
        return False

    def is_in_a_root(self, path, roots):
        """Checks if a directory is physically in one of the supplied roots."""
        for root in roots:
            if self.is_in_root(path, root):
                return True
        return False

    def _update_dir(self, dir, force = False, dispatch_signal = True):
        # dirname/filename : (id, lastmodified)
        library_files = dict([(os.path.join(it[1],it[3]), (it[2],it[4]))\
            for it in self.db_con.get_all_files(dir,self.type)])
        # name : id
        library_dirs = dict([(item[1],item[0]) for item \
                            in self.db_con.get_all_dirs(dir,self.type)])
        # name
        library_dirlinks = [item[1] for item\
                            in self.db_con.get_all_dirlinks(dir, self.type)]

        changes = self.walk_directory(dir or self.get_root_path(),
                library_dirs, library_files, library_dirlinks, force=force,
                dispatch_signal=dispatch_signal)

        # Remove unexistent files and directories from library
        for (id, lastmodified) in library_files.values():
            self.db_con.remove_file(id)
            if dispatch_signal:
                reactor.callFromThread(self.dispatch_signame,\
                        'mediadb.mupdate', attrs = {"type": "remove", "id": id})
            changes.append((id, "remove"))
        for id in library_dirs.values(): self.db_con.remove_dir(id)
        for dirlinkname in library_dirlinks:
            self.db_con.remove_dirlink(dirlinkname, self.type)
            if self.watcher:
                self.watcher.stop_watching_dir(dirlinkname)

        return changes

    def _update(self, force = False):
        self._update_end = False

        try:
            # compare keys recorded in the database with needed key
            # if there is a difference, force update
            keys = self.db_con.get_media_keys(self.search_type)
            # remove cover because it is not used
            try: keys.remove(("cover",))
            except ValueError:
                pass
            if len(keys) > 0 and len(keys) != len(self.media_attr):
                log.msg(\
                    _("%s library has to be updated, this can take a while.")%\
                    (self.type,))
                force = True
            self._update_dir('', force=force)

            self.mutex.acquire()
            self.db_con.erase_empty_dir(self.type)
            self.db_con.update_stats(self.type)
            # commit changes
            self.db_con.connection.commit()
            self.mutex.release()
        finally:
            # close the connection
            self.db_con.close()

    def walk_directory(self, walk_root,
                       library_dirs, library_files, library_dirlinks,
                       force = False, forbidden_roots=None,
                       dispatch_signal=True):
        """Walk a directory for files to update.
        Called recursively to carefully handle symlinks."""
        if not forbidden_roots:
            forbidden_roots = [self.get_root_path()]

        changes = []
        for root, dirs, files in os.walk(walk_root):
            try: root = self._encode(root)
            except UnicodeError: # skip this directory
                continue

            try: dir_id = library_dirs[root]
            except KeyError:
                dir_id = self.db_con.insert_dir(root, self.type)
            else:
                del library_dirs[root]

            # search symlinks
            for dir in dirs:
                try: dir = self._encode(dir)
                except UnicodeError: # skip this directory
                    continue
                # Walk only symlinks that aren't in library root or in one of
                # the additional known root paths which consist in already
                # crawled and out-of-main-root directories
                # (i.e. other symlinks).
                dir_path = os.path.join(root, dir)
                if os.path.islink(dir_path):
                    if not self.is_in_a_root(dir_path, forbidden_roots):
                        forbidden_roots.append(os.path.realpath(dir_path))
                        if dir_path in library_dirlinks:
                            library_dirlinks.remove(dir_path)
                        else:
                            self.db_con.insert_dirlink(dir_path, self.type)
                            if self.watcher:
                                self.watcher.watch_dir(dir_path, self)
                        changes.extend(self.walk_directory(dir_path,
                                 library_dirs, library_files, library_dirlinks,
                                 force, forbidden_roots,
                                 dispatch_signal))

            # else update files
            changes.extend(self.update_files(root, dir_id, files,
                                        library_files, force, dispatch_signal))

        return changes

    def end_update(self, result = True):
        self._update_end = True
        if result: log.msg(_("The %s library has been updated") % self.type)
        else:
            msg = _("Unable to update the %s library. See log.") % self.type
            log.err(msg)
            self._update_error = msg
        return True

    def update_files(self, root, dir_id, files, library_files,
                     force = False, dispatch_signal=True):
        changes = []
        for file in files:
            try: file = self._encode(file)
            except UnicodeError: # skip this file
                continue

            file_path = os.path.join(root, file)
            try:
                fid, lastmodified = library_files[file_path]
                need_update = force or os.stat(file_path).st_mtime>lastmodified
                changes_type = "update"
            except KeyError:
                need_update, fid = True, None
                changes_type = "add"
            else: del library_files[file_path]
            if need_update:
                file_info = self._get_file_info(file_path)
                if file_info is not None: # file supported
                    fid = self.set_media(dir_id, file_path, file_info, fid)
            if fid: self.set_extra_infos(root, file, fid)
            if need_update and fid:
                if dispatch_signal:
                    reactor.callFromThread(self.dispatch_signame,\
                                'mediadb.mupdate',
                                attrs = {"type": changes_type, "id": fid})
                changes.append((fid, changes_type))

        return changes

    def set_media(self, dir_id, file_path, file_info, file_id):
        if file_info is None: return file_id # not supported
        lastmodified = os.stat(file_path).st_mtime
        if file_id: # do not update persistent attribute
            for attr in self.__class__.persistent_attr: del file_info[attr]
            fid = file_id
            self.db_con.update_file(fid, lastmodified)
        else:
            filename = os.path.basename(file_path)
            fid = self.db_con.insert_file(dir_id, filename, lastmodified)
        self.db_con.set_media_infos(fid, file_info)
        return fid

    def set_extra_infos(self, dir, file, file_id):
        pass

    def _get_file_info(self, file_path):
        (base, ext) = os.path.splitext(file_path)
        # try to get infos from this file
        try: file_info = self.ext_dict[ext.lower()]().parse(file_path)
        except (TypeError, KeyError):
            log.info(_("File %s not supported") % file_path)
            return None
        except Exception, ex:
            log.err(_("Unable to get infos from %s, see traceback")%file_path)
            log.err("------------------Traceback lines--------------------")
            log.err(self._encode(traceback.format_exc()))
            log.err("-----------------------------------------------------")
            return None
        return file_info

    #######################################################################
    ## Inotify actions
    #######################################################################
    @inotify_action
    def add_file(self, path, file):
        file_path = os.path.join(path, file)
        file_info = self._get_file_info(file_path)
        if not file_info:
            return self._inotify_add_info(path, file)

        try: dir_id, file_id = self.db_con.is_file_exist(path, file, self.type)
        except TypeError:
            dir_id = self.db_con.is_dir_exist(path, self.type) or\
                     self.db_con.insert_dir(path, self.type)
            file_id = None

        fid = self.set_media(dir_id, file_path, file_info, file_id)
        if fid:
            self.set_extra_infos(path, file, fid)
            self._add_missing_dir(os.path.dirname(path))
            return [(fid, "add")]
        return None

    @inotify_action
    def update_file(self, path, name):
        try: dir_id, file_id = self.db_con.is_file_exist(path, name, self.type)
        except TypeError:
            return self._inotify_update_info(path, name)
        else:
            file_path = os.path.join(path, name)
            file_info = self._get_file_info(file_path)
            if not file_info:
                return self._inotify_update_info(path, name)
            self.set_media(dir_id, file_path, file_info, file_id)
            return [(file_id, "update")]

    @inotify_action
    def remove_file(self, path, name):
        file = self.db_con.is_file_exist(path, name, self.type)
        if file:
            dir_id, file_id = file
            self.db_con.remove_file(file_id)
            self._remove_empty_dir(path)
            return [(file_id, "remove")]
        else:
            return self._inotify_remove_info(path, name)

    @inotify_action
    def add_directory(self, path, name, dirlink=False):
        dir_path = os.path.join(path, name)
        if dirlink:
            self.db_con.insert_dirlink(dir_path, self.type)
            self.watcher.watch_dir(dir_path, self)

        changes = self._update_dir(dir_path.rstrip("/"), dispatch_signal=False)
        self._add_missing_dir(os.path.dirname(dir_path))
        self._remove_empty_dir(path)
        return changes

    @inotify_action
    def remove_directory(self, path, name, dirlink=False):
        dir_path = os.path.join(path, name)
        dir_id = self.db_con.is_dir_exist(dir_path, self.type)
        if not dir_id:
            return None

        if dirlink:
            self.db_con.remove_dirlink(dir_path, self.type)
            self.watcher.stop_watching_dir(dir_path)

        ids = self.db_con.remove_recursive_dir(dir_path)
        self._remove_empty_dir(path)
        return [(id, "remove") for id in ids]

    def inotify_record_changes(self, fids):
        self.mutex.acquire()
        self.db_con.update_stats(self.type)
        self.db_con.connection.commit()
        for fid, signal_type in fids:
            reactor.callFromThread(self.dispatch_signame, 'mediadb.mupdate',\
                    attrs = {"type": signal_type, "id": fid})
        reactor.callFromThread(self.dispatch_signame, self.update_signal_name)
        self.mutex.release()

    def _remove_empty_dir(self, path):
        while path != "":
            if len(self.db_con.get_all_files(path, self.type)) > 0:
                break
            dir_id = self.db_con.is_dir_exist(path, self.type)
            if dir_id: self.db_con.remove_dir(dir_id)
            path = os.path.dirname(path)

    def _add_missing_dir(self, path):
        """ add missing dir in the mediadb """
        while path != "":
            dir_id = self.db_con.is_dir_exist(path, self.type)
            if dir_id: break
            self.db_con.insert_dir(path, self.type)
            path = os.path.dirname(path)


class AudioLibrary(_Library):
    type = "audio"
    search_type = "song"
    update_signal_name = 'mediadb.aupdate'
    custom_attr = ("artist","album","genre","tracknumber","date","bitrate",\
                   "replaygain_track_gain","replaygain_track_peak",\
                   "various_artist","discnumber")
    cover_name = ("cover.jpg", "folder.jpg", ".folder.jpg",\
                  "cover.png", "folder.png", ".folder.png")

    def get_cover(self, file_id):
        try: (cover_id, mime, image) = self.db_con.get_file_cover(file_id)
        except TypeError:
            raise NotFoundException
        return {"mime": mime, "cover": base64.b64decode(image), "id": cover_id}

    def __extract_cover(self, cover_path):
        if os.path.getsize(cover_path) > 512*1024:
            return None # file too large (> 512k)

        # parse video file with kaa
        cover_infos = kaa.metadata.parse(cover_path)
        if cover_infos is None:
            raise TypeError(_("cover %s not supported by kaa parser") % \
                    cover_path)
        # get mime type of this picture
        mime_type = cover_infos["mime"]
        if unicode(mime_type) not in (u"image/jpeg", u"image/png"):
            log.info(_("cover %s : wrong mime type") % cover_path)
            return None

        try: fd = open(cover_path)
        except Exception, ex:
            log.info(_("Unable to open cover file %s") % cover_path)
            return None
        rs = fd.read()
        fd.close()
        return mime_type, base64.b64encode(rs)

    def __find_cover(self, dir):
        cover = None
        for name in self.cover_name:
            cover_path = os.path.join(dir, name)
            if os.path.isfile(cover_path):
                try: (cover, lmod) = self.db_con.is_cover_exist(cover_path)
                except TypeError:
                    try: mime, image = self.__extract_cover(cover_path)
                    except TypeError:
                        return None
                    cover = self.db_con.add_cover(cover_path, mime, image)
                else:
                    if int(lmod)<os.stat(cover_path).st_mtime:
                        try: mime, image = self.__extract_cover(cover_path)
                        except TypeError:
                            return None
                        self.db_con.update_cover(cover, mime, image)
                break
        return cover

    def __update_cover(self, file_id, cover):
        try:
            (recorded_cover, mime, source) = self.db_con.get_file_cover(\
                    file_id,True)
        except TypeError:
            recorded_cover, mime, source = -1, "", ""
        if not source.startswith("hash") and str(recorded_cover) != str(cover):
            self.db_con.set_media_infos(file_id,{"cover": cover})
            return True
        return False

    def __get_digest(self, data):
        return "hash_-_%s" % hashlib.md5(data).hexdigest()

    def _update_dir(self, dir, force=False, dispatch_signal=True):
        changes = super(AudioLibrary, self)._update_dir(dir, force,
                                                        dispatch_signal)
        # remove unused cover
        self.db_con.remove_unused_cover()
        return changes

    def update_files(self, root, dir_id, files, library_files, force=False,
                     dispatch_signal=True):
        changes = []
        if len(files): cover = self.__find_cover(root)
        for file in files:
            try: file = self._encode(file)
            except UnicodeError: # skip this file
                continue

            file_path = os.path.join(root, file)
            try:
                fid, lastmodified = library_files[file_path]
                need_update = force or os.stat(file_path).st_mtime>lastmodified
                changes_type = "update"
            except KeyError:
                need_update, fid = True, None
                changes_type = "add"
            else: del library_files[file_path]
            if need_update:
                file_info = self._get_file_info(file_path)
                if file_info is not None: # file supported
                    fid = self.set_media(dir_id,file_path,file_info,fid,cover)
            elif fid and cover:
                self.__update_cover(fid, cover)
            if need_update and fid:
                if dispatch_signal:
                    reactor.callFromThread(self.dispatch_signame,
                        'mediadb.mupdate',\
                        attrs = {"type": changes_type, "id": fid})
                changes.append((fid, changes_type))
        return changes

    def set_media(self, dir_id, file_path, file_info, file_id, cover = None):
        if file_info is not None and "cover" in file_info:
            # find a cover in the file
            image = base64.b64encode(file_info["cover"]["data"])
            mime = file_info["cover"]["mime"]
            # use hash to identify cover in the db and avoid duplication
            img_hash = self.__get_digest(image)
            try: (cover, lmod) = self.db_con.is_cover_exist(img_hash)
            except TypeError:
                cover = self.db_con.add_cover(img_hash, mime, image)
            file_info["cover"] = cover
        elif cover: # use the cover available in this directory
            file_info["cover"] = cover
        fid = super(AudioLibrary, self).set_media(dir_id, file_path, \
                file_info, file_id)
        # update compilation tag if necessary
        if fid and "album" in file_info.keys() and file_info["album"] != '':
            self.db_con.set_variousartist_tag(fid, file_info)
        return fid

    #
    # custom inotify actions
    #
    @inotify_action
    def add_file(self, path, file):
        file_path = os.path.join(path, file)
        file_info = self._get_file_info(file_path)
        if not file_info and file in self.cover_name: # it is a cover
            files = self.db_con.get_dircontent_id(path, self.type)
            if len(files) > 0:
                file_path = os.path.join(path, file)
                try: mime, image = self.__extract_cover(file_path)
                except TypeError: # image not supported
                    return False
                if image:
                    cover = self.db_con.add_cover(file_path, mime, image)
                    changes = []
                    for (id,) in files:
                        if self.__update_cover(id, cover):
                            changes.append((id, "update"))
                    return changes
            return None

        try: dir_id, file_id = self.db_con.is_file_exist(path, file, self.type)
        except TypeError:
            dir_id = self.db_con.is_dir_exist(path, self.type) or\
                     self.db_con.insert_dir(path, self.type)
            file_id = None

        fid = self.set_media(dir_id, file_path, file_info, file_id)
        if fid:
            self._add_missing_dir(os.path.dirname(path))
            return [(fid, "add")]
        return None

    def _inotify_remove_info(self, path, file):
        rs = self.db_con.is_cover_exist(os.path.join(path, file))
        try: (cover, lmod) = rs
        except TypeError:
            return None
        ids = self.db_con.search_id("cover", cover)
        for (id,) in ids:
            self.db_con.set_media_infos(id, {"cover": ""})
        self.db_con.remove_cover(cover)
        return [(id, "update") for (id,) in ids]

    def _inotify_update_info(self, path, file):
        file_path = os.path.join(path, file)
        rs = self.db_con.is_cover_exist(file_path)
        try: (cover, lmod) = rs
        except TypeError:
            return None
        image = self.__extract_cover(file_path)
        if image: self.db_con.update_cover(cover, image)
        return []
    ###########################################################


class VideoLibrary(_Library):
    type = "video"
    search_type = "video"
    update_signal_name = 'mediadb.vupdate'
    custom_attr = ("videoheight", "videowidth","external_subtitle",\
            "audio_channels", "subtitle_channels")
    subtitle_ext = (".srt",)

    def set_extra_infos(self, dir, file, file_id):
        file_path = os.path.join(dir, file)
        (base_path,ext) = os.path.splitext(file_path)
        sub = ""
        for ext_type in self.subtitle_ext:
            if os.path.isfile(os.path.join(base_path + ext_type)):
                sub = quote_uri(base_path + ext_type)
                break
        try:
            (recorded_sub,) = self.db_con.get_file_info(file_id,\
                                                        "external_subtitle")
        except TypeError:
            recorded_sub = None
        if recorded_sub != sub:
            self.db_con.set_media_infos(file_id,{"external_subtitle": sub})

    #
    # custom inotify actions
    #
    def _inotify_add_info(self, path, file):
        (base_file, ext) = os.path.splitext(file)
        if ext in self.subtitle_ext:
            for video_ext in self.ext_dict.keys():
                try: (dir_id,fid,) = self.db_con.is_file_exist(path,\
                                                base_file+video_ext, self.type)
                except TypeError: pass
                else:
                    uri = quote_uri(os.path.join(path, file))
                    self.db_con.set_media_infos(fid, {"external_subtitle": uri})
                    return [(fid, "update")]
        return None

    def _inotify_remove_info(self, path, file):
        (base_file, ext) = os.path.splitext(file)
        if ext in self.subtitle_ext:
            ids = self.db_con.search_id("external_subtitle",\
                    quote_uri(os.path.join(path, file)))
            for (id,) in ids:
                self.db_con.set_media_infos(id, {"external_subtitle": ""})
            return [(id, "update") for (id,) in ids]
        return None

    def _inotify_update_info(self, path, file):
        return None
    ###########################################################

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from os import path
from threading import local
from deejayd.ui import log

try: from sqlite3 import dbapi2 as sqlite # python 2.5
except ImportError:
    from pysqlite2 import dbapi2 as sqlite
    # Check pysqlite version
    pysqlite_min_version = [2, 2]
    pysqlite_version = map(int, sqlite.version.split('.'))
    if pysqlite_version < pysqlite_min_version:
        sqlite_error=_('This program requires pysqlite version %s or later.')\
            % '.'.join(map(str, pysqlite_min_version))
        log.err(sqlite_error, fatal = True)

LOCK_TIMEOUT = 600
DatabaseError = sqlite.DatabaseError

def str_encode(data):
    if isinstance(data, unicode):
        return data.encode("utf-8")
    return data


class DatabaseWrapper(local):

    def __init__(self, db_file):
        self._file = db_file
        self.connection = None

    def cursor(self):
        if self.connection is None:
            try: self.connection = sqlite.connect(self._file,\
                    timeout=LOCK_TIMEOUT)
            except sqlite.Error:
                error = _("Could not connect to sqlite database %s.")%self._file
                log.err(error, fatal = True)
            # configure connection
            sqlite.register_adapter(str,str_encode)

        return self.connection.cursor(factory = SQLiteCursorWrapper)

    def commit(self):
        if self.connection is not None:
            self.connection.commit()

    def rollback(self):
        if self.connection is not None:
            self.connection.rollback()

    def get_last_insert_id(self, cursor):
        return cursor.lastrowid

    def close(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None


class SQLiteCursorWrapper(sqlite.Cursor):

    def execute(self, query, params=()):
        query = self.convert_query(query, len(params))
        return sqlite.Cursor.execute(self, query, params)

    def executemany(self, query, param_list):
        if len(param_list) == 0: return
        query = self.convert_query(query, len(param_list[0]))
        return sqlite.Cursor.executemany(self, query, param_list)

    def convert_query(self, query, num_params):
        return query % tuple("?" * num_params)


def to_sql(table):
    sql = ["CREATE TABLE %s (" % table.name]
    coldefs = []
    for column in table.columns:
        ctype = column.type.lower()
        if column.auto_increment:
            ctype = "integer PRIMARY KEY"
        elif len(table.key) == 1 and column.name in table.key:
            ctype += " PRIMARY KEY"
        elif ctype == "int":
            ctype = "integer"
        coldefs.append("    %s %s" % (column.name, ctype))
    if len(table.key) > 1:
        coldefs.append("    UNIQUE (%s)" % ','.join(table.key))
    sql.append(',\n'.join(coldefs) + '\n);')
    yield '\n'.join(sql)
    for index in table.indices:
        unique = index.unique and "UNIQUE" or ""
        yield "CREATE %s INDEX %s_%s_idx ON %s (%s);" % (unique,table.name,
              '_'.join(index.columns), table.name, ','.join(index.columns))

custom_queries = [
    # custom indexes
    "CREATE INDEX id_key_value_1x ON media_info(id,ikey,value COLLATE BINARY);",
    "CREATE INDEX id_key_value_2x ON media_info(id,ikey,value COLLATE NOCASE);",
    "CREATE INDEX key_value_1x ON media_info (ikey, value COLLATE BINARY);",
    "CREATE INDEX key_value_2x ON media_info (ikey, value COLLATE NOCASE);",
    # extract from ANALYZE request
    "ANALYZE;",
    "INSERT INTO sqlite_stat1 VALUES('cover', 'cover_source_idx','208 1');",
    "INSERT INTO sqlite_stat1 VALUES('stats',\
            'sqlite_autoindex_stats_1','7 1');",
    "INSERT INTO sqlite_stat1 VALUES('variables',\
            'sqlite_autoindex_variables_1','18 1');",
    "INSERT INTO sqlite_stat1 VALUES('media_info',\
            'key_value_2x','70538 3713 6');",
    "INSERT INTO sqlite_stat1 VALUES('media_info',\
            'key_value_1x','70538 3713 6');",
    "INSERT INTO sqlite_stat1 VALUES('media_info',\
            'id_key_value_2x','70538 16 1 1');",
    "INSERT INTO sqlite_stat1 VALUES('media_info',\
            'id_key_value_1x','70538 16 1 1');",
    "INSERT INTO sqlite_stat1 VALUES('media_info',\
            'sqlite_autoindex_media_info_1','70538 16 1');",
    "INSERT INTO sqlite_stat1 VALUES('media_info',\
            'sqlite_autoindex_media_info_1','70538 16 1');",
    "INSERT INTO sqlite_stat1 VALUES('library',\
            'library_directory_idx','4421 18');",
    "INSERT INTO sqlite_stat1 VALUES('library',\
            'library_name_directory_idx','4421 2 1');",
    "INSERT INTO sqlite_stat1 VALUES('library_dir',\
            'library_dir_name_lib_type_idx','377 2 1');",
    ]

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import os
from deejayd.utils import quote_uri
from deejayd.database import schema

def format_tracknumber(tckn):
    numbers = tckn.split("/")
    try: numbers = ["%02d" % int(num) for num in numbers]
    except (TypeError, ValueError):
        return tckn

    return "/".join(numbers)

def upgrade(cursor, backend, config):
    # get audio/video library
    cursor.execute("SELECT * FROM audio_library;")
    audio_library = cursor.fetchall()
    cursor.execute("SELECT * FROM video_library;")
    video_library = cursor.fetchall()

    # get medialist
    cursor.execute("SELECT DISTINCT name FROM medialist ORDER BY name")
    medialist = {}
    for (name,) in cursor.fetchall():
        if name == "__videocurrent__":
            query = "SELECT l.dir, l.filename\
              FROM medialist p JOIN video_library l \
              ON p.media_id = l.id WHERE p.name = %s ORDER BY p.position"
        else:
            query = "SELECT l.dir, l.filename\
              FROM medialist p JOIN audio_library l\
              ON p.media_id = l.id WHERE p.name = %s ORDER BY p.position"
        cursor.execute(query, (name,))
        medialist[name] = cursor.fetchall()

    # erase old table
    cursor.execute("DROP TABLE audio_library")
    cursor.execute("DROP TABLE video_library")
    cursor.execute("DROP TABLE medialist")

    # create new table/indexes
    new_tables = (
            "library_dir",
            "library",
            "media_info",
            "cover",
            "medialist",
            "medialist_libraryitem",
            "medialist_filters",
            "filters",
            "filters_basicfilters",
            "filters_complexfilters",
            "filters_complexfilters_subfilters",
            )
    for table in schema.db_schema:
        if table.name in new_tables:
            for stmt in backend.to_sql(table):
                cursor.execute(stmt)
    for query in backend.custom_queries:
        cursor.execute(query)

    # set library directory
    audio_dirs, video_dirs = {}, {}
    query = "INSERT INTO library_dir (name,lib_type,type)VALUES(%s,%s,%s)"

    path = config.get("mediadb","music_directory")
    cursor.execute(query, (path.rstrip("/"), "audio", "directory")) # root
    audio_dirs[path.rstrip("/")] = cursor.lastrowid
    for id,dir,fn,type,tit,art,alb,gn,tk,date,len,bt,rpg,rpp in audio_library:
        if type != "file":
            cursor.execute(query, (os.path.join(path,dir,fn), "audio", type))
            audio_dirs[os.path.join(path,dir,fn).rstrip("/")] = cursor.lastrowid

    path = config.get("mediadb","video_directory")
    cursor.execute(query, (path.rstrip("/"), "video", "directory")) # root
    video_dirs[path.rstrip("/")] = cursor.lastrowid
    for id,dir,fn,type,tit,width,height,sub,len in video_library:
        if type != "file":
            cursor.execute(query, (os.path.join(path,dir,fn), "video", type))
            video_dirs[os.path.join(path,dir,fn).rstrip("/")] = cursor.lastrowid

    # set library files
    library_files = {}
    query = "INSERT INTO library (directory,name,lastmodified)VALUES(%s,%s,%s)"
    for id,dir,fn,type,tit,art,alb,gn,tk,date,len,bt,rpg,rpp in audio_library:
        path = config.get("mediadb","music_directory")
        if type == "file":
            try: dir_id = audio_dirs[os.path.join(path,dir).rstrip("/")]
            except KeyError:
                continue
            file_path = os.path.join(path,dir,fn)
            if not os.path.isfile(file_path): continue
            cursor.execute(query, (dir_id,fn,os.stat(file_path).st_mtime))
            file_id = cursor.lastrowid
            infos = {
                "type": "song",
                "filename": fn,
                "uri": quote_uri(file_path),
                "rating": "2",
                "lastplayed": "0",
                "skipcount": "0",
                "playcount": "0",
                "compilation": "0",
                "tracknumber": format_tracknumber(tk),
                "title": tit,
                "genre": gn,
                "artist": art,
                "album": alb,
                "date": date,
                "replaygain_track_gain":rpg,
                "replaygain_track_peak":rpp,
                "bitrate": bt,
                "length": len,
                "cover": "",
                }
            entries = [(file_id, k, v) for k, v in infos.items()]
            cursor.executemany("INSERT INTO media_info\
                (id,ikey,value)VALUES(%s,%s,%s)", entries)
            library_files[os.path.join(path,dir,fn)] = file_id

    for id,dir,fn,type,tit,width,height,sub,len in video_library:
        path = config.get("mediadb","video_directory")
        if type == "file":
            try: dir_id = video_dirs[os.path.join(path,dir).rstrip("/")]
            except KeyError:
                continue
            file_path = os.path.join(path,dir,fn)
            if not os.path.isfile(file_path): continue
            cursor.execute(query, (dir_id,fn,os.stat(file_path).st_mtime))
            file_id = cursor.lastrowid
            infos = {
                "type": "video",
                "filename": fn,
                "uri": quote_uri(file_path),
                "rating": "2",
                "lastplayed": "0",
                "skipcount": "0",
                "playcount": "0",
                "title": tit,
                "length": len,
                "videowidth": width,
                "videoheight": height,
                "external_subtitle": sub,
                }
            entries = [(file_id, k, v) for k, v in infos.items()]
            cursor.executemany("INSERT INTO media_info\
                (id,ikey,value)VALUES(%s,%s,%s)", entries)
            library_files[os.path.join(path,dir,fn)] = file_id

    # set compilation tag
    query = "SELECT DISTINCT value FROM media_info WHERE ikey='album'"
    cursor.execute(query)
    for (album,) in cursor.fetchall():
        if album == '': continue # do not set compilation tag
        query = "SELECT COUNT(DISTINCT m.value) FROM media_info m \
                JOIN media_info m2 ON m.id = m2.id\
                WHERE m.ikey='artist' AND m2.ikey='album'\
                AND m2.value=%s"
        cursor.execute(query, (album,))
        (value,) = cursor.fetchone()

        if int(value) > 1:
            cursor.execute("SELECT id FROM media_info\
                    WHERE ikey='album' AND value=%s", (album,))
            cursor.executemany("UPDATE media_info SET value = '1'\
                    WHERE ikey='compilation' AND id = %s",cursor.fetchall())

    # set medialist
    for pl_name, items in medialist.items():
        cursor.execute("INSERT INTO medialist (name,type)VALUES(%s,%s)",\
                            (pl_name, "static"))
        pl_id = cursor.lastrowid
        path = config.get("mediadb","music_directory")
        if pl_name == "__videocurrent__":
            path = config.get("mediadb","video_directory")
        for item in items:
            try: file_id = library_files[os.path.join(path,item[0],item[1])]
            except KeyError:
                continue
            cursor.execute("INSERT INTO medialist_libraryitem\
                (medialist_id,libraryitem_id)VALUES(%s,%s)", (pl_id, file_id))

    # other updates
    sql = [
        "DELETE FROM variables WHERE name='repeat'",
        "DELETE FROM variables WHERE name='songlistid'",
        "DELETE FROM variables WHERE name='repeat'",
        "DELETE FROM variables WHERE name='random'",
        "DELETE FROM variables WHERE name='qrandom'",
        "INSERT INTO variables VALUES('playlist-playorder','inorder');",
        "INSERT INTO variables VALUES('panel-playorder','inorder');",
        "INSERT INTO variables VALUES('video-playorder','inorder');",
        "INSERT INTO variables VALUES('queue-playorder','inorder');",
        "INSERT INTO variables VALUES('playlist-repeat','0');",
        "INSERT INTO variables VALUES('panel-repeat','0');",
        "INSERT INTO variables VALUES('video-repeat','0');",
        "INSERT INTO variables VALUES('panelid','1');",
        "INSERT INTO variables VALUES('panel-type','panel');",
        "INSERT INTO variables VALUES('panel-value','');",
        "UPDATE variables SET value = '8' WHERE name = 'database_version';",
        ]
    for s in sql:
        cursor.execute(s)

# vim: ts=4 sw=4 expandtab



# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

import time
from threading import local
from deejayd.ui import log
import MySQLdb as mysql

# We want version (1, 2, 1, 'final', 2) or later. We can't just use
# lexicographic ordering in this check because then (1, 2, 1, 'gamma')
# inadvertently passes the version test.
version = mysql.version_info
if (version < (1,2,1) or (version[:3] == (1, 2, 1) and
        (len(version) < 5 or version[3] != 'final' or version[4] < 2))):
            raise ImportError, "MySQLdb-1.2.1p2 or newer is required; you have %s" % mysql.__version__

DatabaseError = mysql.DatabaseError

class DatabaseWrapper(local):
    __slots__ = "globalcommit"

    def __init__(self, db_name, db_user, db_password, db_host, db_port):
        self.connection = None
        self.last_commit = 0

        self.db_name = db_name
        self.db_user = db_user
        self.db_password = db_password
        self.db_host = db_host
        self.db_port = db_port

    def _valid_connection(self):
        if self.connection is not None:
            try:
                self.connection.ping()
                try: globalcommit = self.globalcommit
                except AttributeError:
                    globalcommit = 0
                if self.last_commit < globalcommit:
                    self.connection.commit()
                    self.connection.close()
                    self.connection = None
                    self.last_commit = globalcommit
                else: return True
            except DatabaseError:
                self.connection.close()
                self.connection = None
        return False

    def cursor(self):
        if not self._valid_connection():
            try: self.connection = mysql.connect(db=self.db_name,\
                  user=self.db_user, passwd=self.db_password,\
                  host=self.db_host, port=self.db_port, charset="utf8",\
                  use_unicode=True)
            except DatabaseError, err:
                error = _("Could not connect to MySQL server %s." % err)
                log.err(error, fatal = True)
        cursor = self.connection.cursor()
        return cursor

    def commit(self):
        if self.connection is not None:
            self.connection.commit()
            timestamp = time.time()
            self.globalcommit = timestamp
            self.lastcommit = timestamp

    def rollback(self):
        if self.connection is not None:
            try:
                self.connection.rollback()
            except mysql.NotSupportedError:
                pass

    def get_last_insert_id(self, cursor):
        return cursor.lastrowid

    def close(self):
        if self.connection is not None:
            self.connection.close()
            self.connection = None


def to_sql(table):

    def __collist(table, columns):
        cols = []
        limit = 333 / len(columns)
        if limit > 255:
            limit = 255
        for c in columns:
            name = '`%s`' % c
            table_col = filter((lambda x: x.name == c), table.columns)
            if len(table_col) == 1 and table_col[0].type.lower() == 'text':
                name += '(%s)' % limit
            # For non-text columns, we simply throw away the extra bytes.
            # That could certainly be optimized better, but for now let's KISS.
            cols.append(name)
        return ','.join(cols)

    sql = ['CREATE TABLE %s (' % table.name]
    coldefs = []
    for column in table.columns:
        ctype = column.type
        if ctype == "blob": ctype = "mediumblob"
        if column.auto_increment:
            ctype = 'INT UNSIGNED NOT NULL AUTO_INCREMENT'
            # Override the column type, as a text field cannot
            # use auto_increment.
            column.type = 'int'
        coldefs.append('    `%s` %s' % (column.name, ctype))
    if len(table.key) > 0:
        coldefs.append('    PRIMARY KEY (%s)' %
                       __collist(table, table.key))
    sql.append(',\n'.join(coldefs) + '\n) ENGINE=InnoDB')
    #sql.append(',\n'.join(coldefs) + '\n)')
    yield '\n'.join(sql)

    for index in table.indices:
        unique = index.unique and "UNIQUE" or ""
        yield 'CREATE %s INDEX %s_%s_idx ON %s (%s);' % (unique,table.name,
              '_'.join(index.columns), table.name,
              __collist(table, index.columns))

custom_queries = [
    "CREATE UNIQUE INDEX id_key_value_idx ON media_info\
        (id, ikey(64), value(64));",
    "CREATE INDEX key_value_idx ON media_info\
        (ikey(64), value(64));",
    ]

# vim: ts=4 sw=4 expandtab



#!/usr/bin/env python

# Deejayd, a media player daemon
# Copyright (C) 2007-2009 Mickael Royer <mickael.royer@gmail.com>
#                         Alexandre Rossi <alexandre.rossi@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

"""
This is the test suite launcher :
    * Without arguments, it runs the whole test suite.
    * It accepts a list of arguments which can be :
        - a test module name without the 'test_' prefix.
          e.g. : ./tests.py xmlprocessing
        - a test module name without the 'test_' prefix, a slash, and a test
          name in unittest dotted notation (See the documentation of
          loadTestsFromName at
          http://docs.python.org/lib/testloader-objects.html)
          e.g. : ./tests.py xmlprocessing/TestAnswerParser
              or ./tests.py xmlprocessing/TestAnswerParser.testAnswerParserError
    * If the fist argument is 'list', list all the possibles tests that can be
      combined on the command line restricted by the same arguments.
      e.g. : ./tests.py list xmlprocessing client
      would list all the tests that are to be run from those test modules.
"""

import sys, os, glob
from optparse import OptionParser
import unittest

import testdeejayd

TEST_NAMESPACE = 'testdeejayd'
test_suites_dir = os.path.join(os.path.dirname(__file__), TEST_NAMESPACE)

# Workaround for __import__ behavior, see
# http://docs.python.org/lib/built-in-funcs.html
def my_import(name):
    mod = __import__(name)
    components = name.split('.')
    for comp in components[1:]:
        mod = getattr(mod, comp)
    return mod

def get_testfile_from_id(id):
    return os.path.join(test_suites_dir, "test_%s.py" % id)

def get_id_from_module(module):
    return module.__name__[len(TEST_NAMESPACE+'.'+'test_'):]

def get_all_tests():
    return [(x, None) for x in glob.glob(get_testfile_from_id("*"))]

usage = "usage: %prog [options] [tests-list]"
parser = OptionParser(usage=usage)
parser.add_option("-p","--profile",dest="profile",type="string",\
    help="testserver profile used for this test suite")
parser.set_defaults(profile="default")
(options, myargs) = parser.parse_args()

# update profiles
from testdeejayd import TestCaseWithServer
TestCaseWithServer.profiles = options.profile

tests_to_run = None
list_only = False
args = None
if len(myargs) > 0:
    if myargs[0] == 'list':
        list_only = True
        args = myargs[1:]
    else:
        args = myargs

if args:
    tests_to_consider = []
    for test_id in args:
        try:
            test_module, test_name = test_id.split('/')
        except ValueError:
            test_module = test_id
            test_name = None

        tests_to_consider.append((get_testfile_from_id(test_module), test_name))
else:
    tests_to_consider = get_all_tests()

def get_test_suite(test_module, test_name=None):
    test_suite = None
    if test_name:
        test_suite = unittest.defaultTestLoader.loadTestsFromName(test_name,
                                                                  test_module)
    else:
        test_suite = unittest.defaultTestLoader.loadTestsFromModule(test_module)
    return test_suite

def get_module_and_name(test_id):
    fn, test_name = test_id
    module_path = '.'.join([TEST_NAMESPACE, os.path.basename(fn[:-3])])
    test_module = my_import(module_path)
    return test_module, test_name

def print_tests(class_name, test_name=None):
    if class_name.startswith('Test'):
        has_tests = False
        for fun_name in dir(getattr(test_module, class_name)):
            if fun_name.startswith('test')\
            and (not test_name or fun_name == test_name):
                has_tests = True
                print "%s/%s.%s" % (get_id_from_module(test_module),
                                    class_name, fun_name)
        if has_tests:
            print "%s/%s" % (get_id_from_module(test_module), class_name)

if list_only:
    for test_id in tests_to_consider:
        test_module, test_name = get_module_and_name(test_id)
        if test_name:
            splitted_test_name = test_name.split('.')
            if len(splitted_test_name) > 1:
                class_name, test_name = splitted_test_name
            else:
                class_name, test_name = splitted_test_name[0], None
            print_tests(class_name, test_name)
        else:
            for class_name in dir(test_module):
                print_tests(class_name)
else:
    suitelist = []
    runner = unittest.TextTestRunner(verbosity = 2)
    for test_id in tests_to_consider:
        test_module, test_name = get_module_and_name(test_id)
        suitelist.append(get_test_suite(test_module, test_name))
    runner.run(unittest.TestSuite(suitelist))


# vim: ts=4 sw=4 expandtab
